#include <stdio.h>

#ifdef ORB_ORBACUS
#include <OB/CORBA.h>
#include <OB/Util.h>
#include <ctype.h>
#include <OB/Timer.h>
#else
#include <dispatch/dispatch.h>
#endif

extern "C"
{
#include <lua.h>
#include <lauxlib.h>
}

#include "cotimer.h"

class CoTimer : 
#ifdef ORB_ORBACUS
  public OBTimer
#else
  public IOHandler
#endif
{
   lua_State* myluaState;
   int lref;
   unsigned int t;
   int locked;

   void lock_ref()
   {
      lua_State* L = this->myluaState;
      if (! this->locked)
      {
         int aux_ref = this->lref;
         lua_getref(L, aux_ref);
         this->lref = lua_ref(L, 1);
         this->locked = 1;
         lua_unref(L, aux_ref);
      }
   }

   void unlock_ref()
   {
      lua_State* L = this->myluaState;
      if (this->locked)
      {
         int aux_ref = this->lref;
         lua_getref(L, aux_ref);
         this->lref = lua_ref(L, 0);
         this->locked = 0;
         lua_unref(L, aux_ref);
      }
   }

public:
   CoTimer( lua_State *L, int pos, unsigned long _t)
   {
      this->myluaState = L;  /* save pointer to Lua State */
      lua_pushvalue(L, pos);
      this->lref = lua_ref(L, 0);
      this->t = _t;
      this->locked = 0;
      //Start();
   }

   #ifndef ORB_ORBACUS
   virtual void timerExpired(long,long)
   {
      this->notify();
   }
   #endif

   virtual void notify()
   {
      lua_State *L = this->myluaState;
      int top = lua_gettop(L);

      lua_getref(L, this->lref);
      lua_pushstring(L, "notify");
      lua_gettable(L, -2);
      if (lua_isfunction(L, -1))
      {
	 lua_insert(L, -2);		/* Lua function below self */
         lua_call(L, 1, 1);		/* 1 parameter (self) & 1 result */
         if (! lua_isnil(L, top+1))
         {
            this->Start();
         }
         else
         {
            this->unlock_ref();
         }
      }
      lua_settop(L, top);
   }

   void Start()
   {
      this->lock_ref();

      #ifdef ORB_ORBACUS
      struct timeval tv;
      tv.tv_sec =  t/1000000;
      tv.tv_usec = t%1000000;

      activate(tv,true);
      #else
      Dispatcher::instance().startTimer(t/1000000,t%1000000,this);
      #endif
   }

   virtual void Stop()
   {
      this->unlock_ref();
      #ifdef ORB_ORBACUS
      this->stop();
      #else
      Dispatcher::instance().stopTimer(this);
      #endif
   }

   void setInterval(unsigned int new_t) { this->t = new_t; }
};


static int _RegisterTimer(lua_State *L)
{
   int TIMER_TAG = (int) lua_tonumber(L, -1);   /* upvalue: timer tag */

   unsigned long t = (unsigned long) lua_tonumber(L, 2);
   if (lua_istable(L, 1))
   {
      CoTimer* timer = new CoTimer(L, 1 , t);
      lua_pushusertag(L, timer, TIMER_TAG);
   }
   else
   {
      lua_error(L, "The first parameter is not a table.");
      lua_pushnil(L);
   }
   return 1;
}

static int _TimerStop(lua_State *L)
{
   int TIMER_TAG = (int) lua_tonumber(L, -1);   /* upvalue: timer tag */

   if (lua_tag(L, 1) == TIMER_TAG)
   {
      CoTimer* timer = (CoTimer*) lua_touserdata(L, 1);
      timer->Stop();
   }
   return 0;
}

static int _TimerStart(lua_State *L)
{
   int TIMER_TAG = (int) lua_tonumber(L, -1);   /* upvalue: timer tag */

   if (lua_tag(L, 1) == TIMER_TAG)
   {
      CoTimer* timer = (CoTimer*) lua_touserdata(L, 1);
      timer->Start();
   }
   return 0;
}

static int _TimerInterval(lua_State *L)
{
   int TIMER_TAG = (int) lua_tonumber(L, -1);   /* upvalue: timer tag */

   unsigned long t = (unsigned long) lua_tonumber(L, 2);
   if (lua_tag(L, 1) == TIMER_TAG)
   {
      CoTimer* timer = (CoTimer*) lua_touserdata(L, 1);
      if (timer)
      {
         timer->setInterval(t);
      }
   }
   return 0;
}

static int gc_fb(lua_State *L)
{
   int TIMER_TAG = (int) lua_tonumber(L, -1);   /* upvalue: timer tag */

   if (lua_tag(L, 1) == TIMER_TAG)
   {
      CoTimer* timer = (CoTimer*) lua_touserdata(L, 1);
      if (timer)
      {
         //timer->Stop();
         delete timer;
      }
   }
   return 0;
}

static char timer_code[] =
"function Timer(t)\n"
"   t._ud = RegisterTimer(t,t.interval)\n"
"   t.Start = function (self) TimerStart(self._ud) end\n"
"   t.Stop = function (self) TimerStop(self._ud) end\n"
"   t.setInterval = function (self,t) TimerInterval(self._ud,t) end\n"
"   return t\n"
"end";

static const struct luaL_reg timerfunc[] = {
  {"RegisterTimer", _RegisterTimer},
  {"TimerStart",    _TimerStart},
  {"TimerStop",     _TimerStop},
  {"TimerInterval", _TimerInterval}
};

void cotimer_open(lua_State *L)
{
   // "TIMER TAG" is an upvalue for all Timer functions

   int TIMER_TAG = lua_newtag(L);

   lua_pushnumber(L, TIMER_TAG);
   lua_pushcclosure(L, (lua_CFunction) gc_fb, 1);
   lua_settagmethod(L, TIMER_TAG,"gc");

   for (unsigned int i=0; i<sizeof(timerfunc)/sizeof(timerfunc[0]); i++) {
     lua_pushnumber(L, TIMER_TAG);
     lua_pushcclosure(L, timerfunc[i].func, 1);
     lua_setglobal(L, timerfunc[i].name);
   }

   lua_dostring(L, timer_code);
}

