#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef WIN32
#include <sys/timeb.h>
#else
#include <sys/time.h>
#endif

extern "C"{
#include <lua.h>
#include <lauxlib.h>
}

#include "utils.h"

///////////////////////////////////////////////////////////////////
// Command Line Utilities

static void set_argv(lua_State* L, int argc, char** argv)
{
   lua_newtable(L);

   lua_pushvalue(L, -1);
   lua_setglobal(L, "argv");

   lua_pushstring(L, "n");
   lua_pushnumber(L, (double)argc);
   lua_rawset(L, -3);

   for (int i = 0; i < argc; i++)
   {
      lua_pushstring(L, argv[i]);
      lua_rawseti(L, -2, i+1);
   }
}

static int _dofile(lua_State* L, char* file)
{
   int result = lua_dofile(L, file);
   if (result == 2)
      fprintf(stderr, "luaorb: cannot execute file %s\n", file);
   return result;
}

int utils_proc_params(lua_State *L, int argc, char** argv)
{
   const char* USAGE="Usage: luaorb [options] [file]. Available options are:\n"
                     "  -h or --help  print this message\n"
                     "  -f name       execute file `name'\n"
                     "  -e stat       execute string `stat'\n"
                     "  -i            enter interactive mode\n"
                     "  -s            enter server mode\n"
                     "                This option is valid only if `file' is\n"
                     "                specified or an option `-e' or `-f' is used.\n"
                     "                (not required if option `-i' is used)\n";

   int has_code = 0, has_i = 0, has_s = 0;

   set_argv(L,argc,argv);

   for (int i = 1; i < argc; i++)
   {
      if ((strcmp(argv[i],"-h") == 0) || (strcmp(argv[i],"--help") == 0))
      {
         printf(USAGE);
         exit(0);
      }
      else if (strcmp(argv[i],"-s") == 0)
         has_s = 1;
      else if (strcmp(argv[i],"-i")==0)
         has_i = 1;
      else if (strcmp(argv[i],"-e")==0)
      {
         has_code = 1;
         i++;
         if (i < argc)
            lua_dostring(L,argv[i]);
         else
         {
            fprintf(stderr, USAGE);
            exit(1);
         }
      }
      else if (strcmp(argv[i],"-f")==0)
      {
         has_code = 1;
         i++;
         if (i < argc)
            _dofile(L,argv[i]);
         else
         {
            fprintf(stderr, USAGE);
            exit(1);
         }
      }
      else if (i == (argc-1))
      {
         char* file = argv[i];
         int size = strlen(file);
         if ((size > 4 && strcmp((file+size-4),".lua")==0) ||
             (size > 3 && strcmp((file+size-3),".lo")==0))
         {
            int result = _dofile(L,file);
            if (result == 2)
            {
              fprintf(stderr, "luaorb: cannot execute file %s\n", file);
            }
            else
               has_code = 1;
         }
      }
   }

   int mode = LO_INTERACTIVE_MODE;
   if (has_i || (!has_s && !has_code))
      mode = LO_SERVER_MODE | LO_INTERACTIVE_MODE;
   else if (has_s && has_code)
      mode = LO_SERVER_MODE;
   else if (!has_s)
      mode = 0;
   else
   {
      fprintf(stderr, USAGE);
      exit(1);
   }
   return mode;
}

///////////////////////////////////////////////////////////////////
// Time Utilities

extern "C" {   //Solaris CC
#ifdef WIN32
static int get_time(lua_State *L)
{
   struct _timeb tval;
   _ftime(&tval);
   double t =  ((double)tval.time) + 0.001*tval.millitm;
   lua_pushnumber(L, t);
   return 1;
}
#else
static int get_time(lua_State *L)
{
   struct timeval tval;
   gettimeofday(&tval,NULL);
   double t =  ((double)tval.tv_sec) + 0.000001*tval.tv_usec;
   lua_pushnumber(L, t);
   return 1;
}
#endif
}

/////////////////////////////////////////////////////////////////////
// Process Utilities

#ifdef WIN32
#include <windows.h>

/*
** Lua table at the top of the stack
*/

static void fill_env(lua_State *L,  char* buffer)
{
   int n = lua_getn(L, -1);	// n = size of "env" table
   int j = 0;

   for( int i = 1; i <= n ; i++)
   {
      lua_pushnumber(L, (double)i);
      lua_gettable(L, -2);
      const char* str = lua_tostring(L, -1);
      strcpy(&buffer[j],str);
      j += (strlen(str) +1 );
      lua_pop(L, 1);
   }
}

static WORD getShowMode(const char* str_mode)
{
   if (!strcmp("SW_SHOWMAXIMIZED", str_mode))
      return SW_SHOWMAXIMIZED;
   else if (!strcmp("SW_SHOWMINIMIZED", str_mode))
      return SW_SHOWMINIMIZED;
   else if (!strcmp("SW_SHOWMINNOACTIVE", str_mode))
      return SW_SHOWMINNOACTIVE;
   else if (!strcmp("SW_RESTORE", str_mode))
      return SW_RESTORE;
   else
      return SW_SHOWNORMAL;
}

static int _create_process (lua_State *L)
{
   const char* commandline;
   WORD show_mode = SW_SHOWNORMAL;
   WORD flags = NORMAL_PRIORITY_CLASS;
   const char* dir = NULL;
   char* env = NULL;
   static char buffer[1024];

   int top = lua_gettop(L);

   luaL_arg_check(L, lua_istable(L,1) || lua_isstring(L,1),1,"must be a table or string");
   if (lua_isstring(L,1))
      commandline = lua_tostring(L,1);
   else
   {
      lua_pushstring(L, "cmd");
      lua_gettable(L, 1);
      if (lua_isstring(L, -1))
        commandline = lua_tostring(L, -1);
      else
         lua_error(L, "The cmd field must be provided");

      lua_pushstring(L, "console");
      lua_gettable(L, 1);
      if (!lua_isnil(L, -1))
         flags |= CREATE_NEW_CONSOLE;

      lua_pushstring(L, "show_mode");
      lua_gettable(L, 1);
      if (lua_isstring(L, -1))
         show_mode = getShowMode(lua_tostring(L, -1));

      lua_pushstring(L, "dir");
      lua_gettable(L, 1);
      if (lua_isstring(L, -1))
         dir = lua_tostring(L, -1);

      lua_pushstring(L, "env");
      lua_gettable(L, 1);
      if (lua_istable(L, -1))
      {
         fill_env(L, buffer);
         env = buffer;
      }
   }

   STARTUPINFO si;
   PROCESS_INFORMATION pi;
   ZeroMemory(&si, sizeof(STARTUPINFO));
   si.cb = sizeof(STARTUPINFO);
   si.dwFlags = STARTF_USESHOWWINDOW;
   si.wShowWindow = show_mode;
   BOOL pr = CreateProcess(NULL,(char*)commandline,NULL,NULL,FALSE,
                           flags,env,dir,&si,&pi);
   CloseHandle(pi.hThread);
   CloseHandle(pi.hProcess);
   if (!pr)
      lua_pushnil(L);
   else
      lua_pushnumber(L,pi.dwProcessId);
   return 1;
}

static int _kill_process(lua_State *L)
{
   DWORD pid = (DWORD)luaL_check_number(L,1);
   HANDLE hProcess = OpenProcess(PROCESS_TERMINATE,0,pid);

   if (hProcess != NULL)
   {
      TerminateProcess(hProcess, 0);
      CloseHandle(hProcess);
   }
   return 0;
}

static int _sleep(lua_State* L)
{
   double dtime = luaL_check_number(L,1);
   DWORD time = (DWORD) (dtime*1000);
   Sleep(time);
   return 0;
}

static void init_proc_utils(lua_State *L)
{
   lua_register(L, "create_process",_create_process);
   lua_register(L, "kill_process",_kill_process);
   lua_register(L, "sleep",_sleep);
}
#else

#include <unistd.h>

static int _sleep(lua_State* L)
{
   double dtime = luaL_check_number(L,1);
   unsigned long time = (unsigned long) (dtime*1000000);
   usleep(time);
   return 0;
}

static void init_proc_utils(lua_State *L)
{
   char proc_code[] =
      "function create_process(p)\n"
      "   local cmd\n"
      "   if type(p) == 'string' then\n"
      "      cmd = p\n"
      "   else\n"
      "      cmd = p.cmd\n"
      "      if type(p.dir) == 'string' then\n"
      "         cmd = format('cd %s; %s', p.dir, cmd)\n"
      "      end\n"
      "   end\n"
      "   local file = tmpname()\n"
      "   local shell_cmd = format(\"tcsh -c '%s & ; echo $! > %s'\", cmd, file)\n"
      "   execute(shell_cmd)\n"
      "   readfrom(file)\n"
      "   local pid = read()\n"
      "   readfrom()\n"
      "   remove(file)\n"
      "   return pid\n"
      " end\n"
      "\n"
      "function kill_process(pid)\n"
      "   execute('kill -9 ' .. pid)\n"
      "end\n";
   lua_dostring(L, proc_code);
   lua_register(L, "sleep",_sleep);
}
#endif

/////////////////////////////////////////////////////////////////////
// Library Initialization


static char util_code[] =
"-----------------------------------------------------------------------\n"
"-- Function to read an IR in a specific file\n"
"-----------------------------------------------------------------------\n"
"function readIOR(file)\n"
"   if readfrom(file) then\n"
"      local ior = read()\n"
"      readfrom()\n"
"      return ior\n"
"   else\n"
"      return nil\n"
"   end\n"
"end\n"
"LUAORB_DIR = getenv('LUAORB_DIR')\n"
"if not LUAORB_DIR then LUAORB_DIR = '.' end\n"
"function include(file)\n"
"   dofile(LUAORB_DIR .. '/lua/' .. file)\n"
"end\n"
"include('env.lua')\n";


void utils_open(lua_State *L)
{
   lua_register(L,"get_time",get_time);
   lua_dostring(L, util_code);
   init_proc_utils(L);
}

