//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_cfg.cpp,v 1.15.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************

#include <stdio.h>
#include <stdarg.h>

extern "C"
{
#include <lua.h>
#include <lauxlib.h>
}

#include "orb_inc.h"
#include "lo_cfg.h"
#include "lo_proxy.h"
#include "lo_except.h"

#ifndef WIN32
#include <unistd.h> //gethostname
#include <netdb.h> //gethostbyname
#endif

/*
** User-configurable modes  (keys used in Lua global table LuaOrbCfg)
*/

static const char* cfg_mode[] = {"verbose_mode",
			         "enum_as_numbers", 
				 "servant_optimization",
				 "OctetSeq_as_string"};

/*
** Registry Data Keys
*/
const char* LuaOrbCfg::reg_NewServantTagTable = "LuaOrb_NewServantTable";
const char* LuaOrbCfg::reg_LuaOrbCfg = "LuaOrbCfg";


/****************************************************************************/
/* LuaOrbCfg Lua methods						    */
/****************************************************************************/

static int _lo_settagmethod(lua_State *L)
{
   int tag = luaL_check_int(L, 1);
   const char* method_name = luaL_check_string(L, 2);

   if (strcmp("newservant",method_name) == 0)
      LuaOrbCfg::setNewServantTagMethod(L, 3, tag); 
   else
      luaL_verror(L,"LuaOrb: Invalid Event %s for LuaOrbCfg.settagmethod.",
		  method_name);

   return 1;  // previous tag method is at the top of the stack
}

static int _lo_gettagmethod(lua_State *L)
{
   int tag = luaL_check_int(L, 1);
   const char* method_name = luaL_check_string(L, 2);

   if (strcmp("newservant",method_name) == 0)
      LuaOrbCfg::getNewServantTagMethod(L, tag);
   else
      luaL_verror(L, "LuaOrb: Invalid Event %s for LuaOrbCfg.gettagmethod.",
                   method_name);

   return 1;  // tag method is at the top of the stack
}

static int _lo_setIFR(lua_State *L)
{
   LuaOrbCfg* locfg = LuaOrbCfg::getConfiguration(L);

   CORBA::Object_ptr obj = CORBA::Object::_nil();

#if defined (ORB_VISIBROKER) || defined (ORB_MICO)
   if (lua_isstring(L, 1) && strncmp("name:",lua_tostring(L, 1),5)==0)
   {
      char* name = lua_tostring(L, 1);
      name = &name[5];
      obj = locfg->orb->bind("CORBA::Repository",name);
   }
   else
#endif
   if (lua_isstring(L, 1))
   {
      try
      {
         obj = locfg->orb->string_to_object(lua_tostring(L, 1));
      }
      catch (CORBA::Exception& e)
      {
         luaL_verror(L, "Invalid IOR: %s",exception_name(e));
      }
   }
   else
   {
      LuaProxy* proxy  = LuaProxy::from_lua(L, 1);
      if (proxy)
         obj = proxy->orb_object;
   }

   if (CORBA::is_nil(obj) && !lua_isnil(L, 1))
      lua_error(L, "LuaOrb Error: Invalid Interface Repository Reference.");

   if (locfg->orb_ir)
   {
      CORBA_release(locfg->orb_ir);
      locfg->orb_ir = CORBA::Repository::_nil();
   }

   try
   {
      if (!CORBA::is_nil(obj))
         locfg->orb_ir = CORBA::Repository::_narrow(obj);
   }
   catch (CORBA::Exception& e)
   {
      luaL_verror(L,"Invalid Object Reference: %s",exception_name(e));
   }

   return 0;
}

static int _lo_getIFR(lua_State *L)
{
   LuaOrbCfg* locfg = LuaOrbCfg::getConfiguration(L);

   if (CORBA::is_nil(locfg->orb_ir))
      lua_pushnil(L);
   else
   {
      try
      {
         LuaProxy* proxy = new LuaProxy(L, "CORBA::Repository",
                                  CORBA::Object::_duplicate(locfg->orb_ir));
         LuaProxy::push(L, proxy);
      }
      catch (Err_General& excep)
      {
         lua_error(L, excep.msg);
      }
   }
  return 1;
}

static int _lo_getHostname(lua_State *L)
{
   char hostname[100];

   if (!gethostname(hostname,100))
   {
      const struct hostent* host_info = gethostbyname(hostname);
      if (host_info != NULL)
         lua_pushstring(L, host_info->h_name);
      else
         lua_pushnil(L);
   }
   else
      lua_pushnil(L);
   return 1;
}

/****************************************************************************/
/* Access to LuaOrb Configuration					    */
/****************************************************************************/

LuaOrbCfg* LuaOrbCfg::getConfiguration(lua_State *L)
{
   lua_getregistry(L);
   lua_pushstring(L, reg_LuaOrbCfg);
   lua_gettable(L, -2);
   LuaOrbCfg* locfg = (LuaOrbCfg*) lua_touserdata(L, -1);

   if (locfg == NULL)
      lua_error(L, "LuaOrb configuration not found (state closed?)");
   lua_pop(L, 2);
   return locfg;
}

// Mode configurations obtained from the Lua global table LuaOrbCfg
int LuaOrbCfg::modeConfigured(lua_State *L, ConfigurableMode mode)
{
   lua_getglobal(L, "LuaOrbCfg");
   lua_pushstring(L, cfg_mode[mode]);
   lua_gettable(L, -2);
   int configured = !lua_isnil(L, -1);
   lua_pop(L, 2);
   return configured;
}

/****************************************************************************/
/* LuaOrbCfg Constructors and Destructor				    */
/****************************************************************************/

static CORBA::Repository_ptr get_ir_ref(CORBA::ORB_ptr orb)
{
   CORBA::Object_var obj;
   try {
      obj = orb->resolve_initial_references("InterfaceRepository");
   }
   catch (...)
   {
      obj = CORBA::Object::_nil();
   }

   if (!CORBA::is_nil(obj))
      return CORBA::Repository::_narrow(obj);
   else
      return CORBA::Repository::_nil();
}

#ifdef CORBA_2_3
LuaOrbCfg::LuaOrbCfg(lua_State *L, CORBA::ORB_ptr _orb,
		     PortableServer::POA_ptr root_poa, const char* name)
{
   char POA_name[256];

   this->orb = _orb;
   this->orb_ir = get_ir_ref(_orb);
   this->state_name = strdup(name);

#ifndef ORB_TAO
   CORBA::Object_var obj = _orb->resolve_initial_references("DynAnyFactory");
   if (CORBA::is_nil(obj))
      lua_error(L, "LuaOrb Initialization Error: DynAny Factory could not be created.");

   this->orb_dynany_factory = DynamicAny::DynAnyFactory::_narrow(obj);

   if (CORBA::is_nil(this->orb_dynany_factory))
      lua_error(L, "LuaOrb Initialization Error: DynAny Factory could not be created.");
#endif

   PortableServer::POAManager_var manager = root_poa->the_POAManager();

   sprintf(POA_name,"%sPOA",this->state_name);
   CORBA::PolicyList plist;
   this->poa = root_poa->create_POA(POA_name, manager, plist);

   try {
      CORBA::PolicyList plist2(2);
      plist2.length(2);
      plist2[0] = root_poa->create_lifespan_policy(PortableServer::PERSISTENT);
      plist2[1] = root_poa->create_id_assignment_policy(PortableServer::USER_ID);

      sprintf(POA_name,"%sPersistentPOA",this->state_name);
      this->persistent_poa = root_poa->create_POA(POA_name,manager,plist2);
   }
   catch (PortableServer::POA::InvalidPolicy&)
   {
      this->persistent_poa = NULL;
   }

}
#else
LuaOrbCfg::LuaOrbCfg(lua_State *L, CORBA::ORB_ptr _orb,
		     CORBA::BOA_ptr boa, const char* name)
{
   this->orb = _orb;
   this->orb_ir = get_ir_ref(_orb);
   this->boa = boa;
   this->state_name = strdup(name);
}
#endif

LuaOrbCfg::~LuaOrbCfg()
{
    if (this->state_name)
       free(this->state_name);

    if (this->orb_ir)
       CORBA_release(this->orb_ir);

#ifdef CORBA_2_3
#ifndef ORB_TAO
    if (this->orb_dynany_factory)
       CORBA_release(this->orb_dynany_factory);
#endif

    CORBA::release(this->poa);
    CORBA::release(this->persistent_poa);
#endif

}

#ifdef CORBA_2_3
void LuaOrbCfg::open(lua_State *L, CORBA::ORB_ptr _orb, 
		     PortableServer::POA_ptr root_poa, const char* state_name)
#else
void LuaOrbCfg::open(lua_State *L, CORBA::ORB_ptr _orb, CORBA::BOA_ptr boa,
		     const char* state_name)
#endif
{
    /* if a state name was not provided, use default */
    if (state_name == NULL)
       state_name = "LuaOrb";

   // Create the Lua global table that stores LuaOrb configuration
   lua_newtable(L);

   lua_pushstring(L, cfg_mode[SERVANT_OPTIMIZATION]);
   lua_pushnumber(L, 1.0);
   lua_rawset(L, -3);
   lua_pushstring(L, cfg_mode[OCTETSEQ_AS_STRING]);
   lua_pushnumber(L, 1.0);
   lua_rawset(L, -3);

   lua_pushstring(L, "settagmethod");
   lua_pushcfunction(L, (lua_CFunction) _lo_settagmethod);
   lua_rawset(L, -3);
   lua_pushstring(L, "gettagmethod");
   lua_pushcfunction(L, (lua_CFunction) _lo_gettagmethod);
   lua_rawset(L, -3);
   lua_pushstring(L, "getIFR");
   lua_pushcfunction(L, (lua_CFunction)_lo_getIFR);
   lua_rawset(L, -3);
   lua_pushstring(L, "setIFR");
   lua_pushcfunction(L, (lua_CFunction)_lo_setIFR);
   lua_rawset(L, -3);
   lua_pushstring(L, "getHostname");
   lua_pushcfunction(L, (lua_CFunction)_lo_getHostname);
   lua_rawset(L, -3);

   lua_pushstring(L, "name");
   lua_pushstring(L, state_name);
   lua_rawset(L, -3);

   lua_setglobal(L, "LuaOrbCfg");

   // Create the LuaOrb tag methods table and store it in the Lua Registry
   lua_getregistry(L);
   lua_pushstring(L, reg_NewServantTagTable);
   lua_newtable(L);
   lua_settable(L, -3);

   // Store LuaOrb configuration in the Registry
#ifdef CORBA_2_3
   LuaOrbCfg* cfg = new LuaOrbCfg(L,_orb,root_poa,state_name);
#else
   LuaOrbCfg* cfg = new LuaOrbCfg(L,_orb,boa,state_name);
#endif
   lua_pushstring(L, reg_LuaOrbCfg);
   lua_pushuserdata(L, cfg);
   lua_settable(L, -3);

   lua_pop(L, 1);

}

void LuaOrbCfg::close(lua_State *L)
{
   int top = lua_gettop(L);

   lua_getregistry(L);
   lua_pushstring(L, reg_NewServantTagTable);
   lua_pushnil(L);
   lua_settable(L, -3);

   lua_pushstring(L, reg_LuaOrbCfg);
   lua_gettable(L, -2);
   LuaOrbCfg* locfg = (LuaOrbCfg*) lua_touserdata(L, -1);
   if (locfg)
   {
      delete locfg;
      lua_pushstring(L, reg_LuaOrbCfg);
      lua_pushnil(L);
      lua_settable(L, -4);	/* previous value still above registry! */
   }
   lua_settop(L, top);
}

/****************************************************************************/
/* LuaOrb Tag Methods							    */
/****************************************************************************/

// getNewServantTagMethod returns the tag method at the top of the stack

void LuaOrbCfg::getNewServantTagMethod(lua_State *L, int tag)
{
   lua_getregistry(L);
   lua_pushstring(L, reg_NewServantTagTable);
   lua_gettable(L, -2);
   lua_pushnumber(L, (double)tag);
   lua_gettable(L, -2);

   lua_insert(L, -3);			// insert tag method below registry
   lua_pop(L, 2);			// remove registry and tag table
}

// setNewServantTagMethod receives the stack position of the new tag
// method (pos) and returns the previous tag method at the top
// of the stack
 
void LuaOrbCfg::setNewServantTagMethod(lua_State *L, int pos, int tag)
{
   if (lua_isfunction(L, pos) || lua_isnil(L, pos))
   {
      // get previous tag method
      lua_getregistry(L);
      lua_pushstring(L, reg_NewServantTagTable);
      lua_gettable(L, -2);
      lua_pushnumber(L, (double)tag);
      lua_gettable(L, -2);		// previous tag method
      lua_insert(L, -3);		// insert it below registry

      // set new tag method
      lua_pushnumber(L, (double)tag);
      lua_pushvalue(L, pos);
      lua_settable(L, -3);

      lua_pop(L, 2);			// remove registry and tag table
   }
   else
      lua_error(L, "LuaOrb: Invalid Tag Method (a tag method must be a function)");
}
