//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_class.cpp,v 1.16.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <orb_inc.h>

extern "C"
{
#include <lua.h>
}

#include <luaorb.h>
#include "lo_class.h"
#include "lo_cfg.h"
#include "lo_except.h"
#include "lo_map.h"
#include "lo_util.h"

//**************************** Class LuaOrbClass ****************************

const char* LuaOrbClass::reg_CLASSES = "LuaOrb_Classes";

LuaOrbClass* LuaOrbClass::getFirstClass(lua_State *L)
{
   LuaOrbClass* firstClass = NULL;

   lua_getregistry(L);
   lua_pushstring(L, LuaOrbClass::reg_CLASSES);
   lua_gettable(L, -2);
   if (!lua_isnil(L, -1))
      firstClass = (LuaOrbClass*) lua_touserdata(L, -1);
   lua_pop(L, 2);
   return firstClass;
}

void LuaOrbClass::setFirstClass(lua_State *L, LuaOrbClass* loClass)
{
   lua_getregistry(L);
   lua_pushstring(L, LuaOrbClass::reg_CLASSES);

   if (loClass == NULL)
      lua_pushnil(L);
   else
      lua_pushuserdata(L, loClass);
   lua_settable(L, -3);
   lua_pop(L, 1);
}

void LuaOrbClass::close(lua_State *L)
{
   LuaOrbClass* next = LuaOrbClass::getFirstClass(L);
   LuaOrbClass* current = next;

   while (current)
   {
       next = current->next;
       delete current;
       current = next;
   }
   LuaOrbClass::setFirstClass(L, NULL);
}

LuaOrbClass* LuaOrbClass::get_class_by_obj(lua_State *L, CORBA::Object_ptr obj)
{
   LuaOrbClass* cl = NULL;
   try
   {
      CORBA::InterfaceDef_ptr idef = obj->_get_interface();
      if (!CORBA::is_nil(idef))
      {
         cl = LuaOrbClass::get_class_by_interfacedef(L, idef);
      }
   }
   catch(...) {}

   if (cl == NULL)
   {
      char* interface_id = LuaOrbUtil::getRepositoryId(L, obj);
      if (interface_id)
      {
         cl = LuaOrbClass::get_class_by_id(L, interface_id);
         free(interface_id);
      }
   }

   if (cl == NULL)
         throw Err_General("LuaOrb: Interface description not available");

   return cl;
}

LuaOrbClass* LuaOrbClass::get_class_by_id(lua_State *L, const char* id)
{
   LuaOrbClass* firstClass = LuaOrbClass::getFirstClass(L);
   LuaOrbClass* current = firstClass;

   while (current)
   {
      if (current->_repository_id &&
          strcmp(id, current->_repository_id) == 0)
         return current;
      current = current->next;
   }
   if (current == NULL)
   {
      current = new LuaOrbClass;
      current->repository_id((char*)id);
      try {
         current->init(L);
      }
      catch (...) {
         delete current;
         throw;
      }
      current->next = firstClass;
      LuaOrbClass::setFirstClass(L, current);
   }
   return current;
}

LuaOrbClass* LuaOrbClass::get_class_by_name(lua_State *L, const char* iname)
{
   LuaOrbClass* firstClass = LuaOrbClass::getFirstClass(L);
   LuaOrbClass* current = firstClass;

   while (current)
   {
      if (current->_interface_name &&
          strcmp(iname, current->_interface_name) == 0)
         return current;
      current = current->next;
   }
   if (current == NULL)
   {
      current = new LuaOrbClass;
      current->interface_name((char*)iname);
      try {
         current->init(L);
      }
      catch (...) {
         delete current;
         throw;
      }
      current->next = firstClass;
      LuaOrbClass::setFirstClass(L, current);
   }
   return current;
}

LuaOrbClass* LuaOrbClass::get_class_by_interfacedef(lua_State *L, 
                                                 CORBA::InterfaceDef_ptr idef)
{
   CORBA::String_var id = idef->id();
   LuaOrbClass* firstClass = LuaOrbClass::getFirstClass(L);
   LuaOrbClass* current = firstClass;

   while (current)
   {
      if (current->_repository_id &&
          strcmp(id, current->_repository_id) == 0)
         return current;
      current = current->next;
   }

   if (current == NULL)
   {
      current = new LuaOrbClass;
      current->_interfacedef = CORBA::InterfaceDef::_duplicate(idef);
      current->repository_id((char*)id);
      try {
         current->init(L);
      }
      catch (...) {
         delete current;
         throw;
      }
      current->next = firstClass;
      LuaOrbClass::setFirstClass(L, current);
   }

   return current;
}

void LuaOrbClass::init(lua_State *L)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);

   char* rep_id = this->repository_id();
   char* itf_name = this->interface_name();

   if (CORBA::is_nil(this->_interfacedef))
   {
      CORBA::Contained_ptr contained = NULL;

      //The Interface Repository is required here
      if (CORBA::is_nil(loCfg->orb_ir))
         throw Err_General("LuaOrb: No Interface Repository Available");

      if (rep_id)
         contained = loCfg->orb_ir->lookup_id(rep_id);
      else if (itf_name)
         contained = loCfg->orb_ir->lookup(itf_name);

      if (contained)
      {
         this->_interfacedef = CORBA::InterfaceDef::_narrow(contained);
         CORBA_release(contained);
      }
   }

   if (!CORBA::is_nil(this->_interfacedef))
   {
      if (rep_id == NULL)
      {
         CORBA::String_var id = this->_interfacedef->id();
         this->repository_id(id);
      }
      if (itf_name == NULL)
      {
         CORBA::String_var name = this->_interfacedef->absolute_name();
         this->interface_name(name);
      }
   }
   else if ((rep_id && strcmp("IDL:omg.org/CORBA/Object:1.0",rep_id)==0) ||
            (itf_name && strcmp("CORBA::Object",itf_name)==0))
   {
      if (rep_id == NULL)
         this->repository_id((char*)"IDL:omg.org/CORBA/Object:1.0");
      if (itf_name == NULL)
         this->interface_name((char*)"CORBA::Object");
   }
   else if ((rep_id && strcmp("IDL:omg.org/CORBA/Repository:1.0",rep_id)==0) ||
            (itf_name && strcmp("CORBA::Repository",itf_name)==0))
   {
      if (rep_id == NULL)
         this->repository_id((char*)"IDL:omg.org/CORBA/Repository:1.0");
      if (itf_name == NULL)
         this->interface_name((char*)"CORBA::Repository");
   }
   else
   {
      char msg[100];
      char* interface_name = (itf_name) ? itf_name : rep_id;

      if (interface_name)
         sprintf(msg,"The Interface %s does not exist in the current Interface Repository.",
                 interface_name);
      else
         sprintf(msg,"Invalid Interface");
      throw Err_General(msg);
   }
}

CORBA::InterfaceDef_ptr LuaOrbClass::get_interfacedef()
{
   return this->_interfacedef;
}

MethodDescr* LuaOrbClass::get_opdef(lua_State *L, char* name)
{
   MethodDescr* descr = this->method_cache->get(name);

   if (!descr)
   {
      //the cache does not have this method description
      CORBA::InterfaceDef_ptr itf_def = this->get_interfacedef();

      if (!itf_def)
      {
         // If the class object does not have a CORBA::InterfaceDef,
         // try to get one from the IFR.
         this->init(L);
         itf_def = this->get_interfacedef();
      }

      if (itf_def)
      {
         CORBA::ContainedSeq* seq = itf_def->lookup_name(name,-1,CORBA(dk_Operation),0);
         if (seq->length() == 1)
         {
            CORBA::OperationDef_ptr opdef=CORBA::OperationDef::_narrow((*seq)[0]);
            if (opdef)
            {
               descr = this->add_opdef(name,opdef);
               CORBA_release(opdef);
            }
         }
         CORBA_release(seq);
      }
   }
   return descr;
}

MethodDescr* LuaOrbClass::add_opdef(char* name, CORBA::OperationDef_ptr def)
{
   MethodDescr* descr = new MethodDescr;
   descr->name = CORBA::string_dup(name);

   CORBA::Contained::Description* orb_descr = def->describe();

   #if defined(CORBA_2_3) && !defined(ORB_TAO)
   const CORBA::OperationDescription* aux_ptr;
   #else
   CORBA::OperationDescription* aux_ptr;
   #endif

   orb_descr->value >>= aux_ptr;
   descr->orb_descr = new CORBA::OperationDescription(*aux_ptr);

   this->method_cache->insert(descr);
   delete orb_descr;
   return descr;
}

AttribDescr* LuaOrbClass::get_attdef(char* name)
{
   AttribDescr* descr = this->attr_cache->get(name);

   if (!descr)
   {
      //the cache does not have this attribute description
      CORBA::InterfaceDef_ptr itf_def = this->get_interfacedef();
      if (itf_def)
      {
         CORBA::ContainedSeq* seq = itf_def->lookup_name(name,1,CORBA(dk_Attribute),0);
         if (seq->length() == 1)
         {
            CORBA::AttributeDef_ptr attdef = CORBA::AttributeDef::_narrow((*seq)[0]);
            if (attdef)
            {
               descr = this->add_attdef(name,attdef);
               CORBA_release(attdef);
            }
         }
         CORBA_release(seq);
      }
   }
   return descr;
}

AttribDescr* LuaOrbClass::add_attdef(char* name, CORBA::AttributeDef_ptr def)
{
   AttribDescr* descr = new AttribDescr;
   descr->name = CORBA::string_dup(name);

   CORBA::Contained::Description* orb_descr = def->describe();

   #if defined(CORBA_2_3) && !defined(ORB_TAO)
   const CORBA::AttributeDescription* aux_ptr;
   #else
   CORBA::AttributeDescription* aux_ptr;
   #endif

   orb_descr->value >>= aux_ptr;
   descr->orb_descr = new CORBA::AttributeDescription(*aux_ptr);

   this->attr_cache->insert(descr);
   delete orb_descr;
   return descr;
}

//**************************** Class MethodDescr *********************************

/*
** Parameters to be marshalled are in the Lua Stack
**   ( "nparams" values, beginning at stack position "pos")
*/
void MethodDescr::marshal_params(lua_State *L, int pos, unsigned int nparams,
                                 CORBA::NVList_ptr arguments)
{
    unsigned int in_params = 0;

    for (unsigned int i=0;i<this->orb_descr->parameters.length();i++)
    {
       CORBA::Flags mode;
       CORBA::ParameterDescription& param = this->orb_descr->parameters[i];
       if ( param.mode == CORBA(PARAM_IN))
           mode = CORBA(ARG_IN);
       else if ( param.mode == CORBA(PARAM_OUT))
           mode = CORBA(ARG_OUT);
       else if ( param.mode == CORBA(PARAM_INOUT))
           mode = CORBA(ARG_INOUT);
       try
       {
        if ( mode == CORBA(ARG_IN) || mode == CORBA(ARG_INOUT))
        {
           if (in_params >= nparams)
              throw Err_NParams(this->orb_descr->parameters.length(),
                                nparams);

           CORBA::Any_var anyValue = lua2any(L, pos, param.type);
#ifdef ORB_TAO
           arguments->add_value(param.name,anyValue.in(),mode);
#else
           arguments->add_value(param.name,anyValue,mode);
#endif
           in_params++;
	   pos++;	// advance stack position
        }
        else
        {
           CORBA::NamedValue_ptr nv = arguments->add_item(param.name,mode);
#ifdef ORB_TAO
           nv->value()->type(param.type);
#else
           nv->value()->replace(param.type,NULL);
#endif
        }
       }
       catch (Err_Param& ex)
       {
        ex.index = in_params+1;
        throw;
       }
    }

}

CORBA::Any* MethodDescr::mount_result(CORBA::Request_ptr req)
{
   if(this->orb_descr->result &&
      this->orb_descr->result->kind() != CORBA(tk_void))
   {
      CORBA::Any* resultAny = NULL;
      CORBA::NamedValue* result = req->result();
      resultAny = result->value();

   #ifdef CORBA_2_3
      req->set_return_type(this->orb_descr->result);
   #else
      resultAny->replace(this->orb_descr->result, &result);
   #endif

      return resultAny;
   }
   else
      return NULL;
}

/*
** Unmarhalled results are returned in the Lua stack
**   (the number of results pushed in the stack is returned) 
*/
unsigned int MethodDescr::unmarshal_result(lua_State *L, CORBA::Any* resultAny, 
					   CORBA::NVList_ptr arguments)
{
   unsigned int nresults = 0;

   //maps the result value
   if (resultAny)
   {
      any2lua(L, this->orb_descr->result, resultAny);
      nresults++;
   }

    //maps the out params
    for (unsigned int i=0;i<this->orb_descr->parameters.length();i++)
    {
       CORBA::ParameterDescription& param = this->orb_descr->parameters[i];
       if ( param.mode == CORBA(PARAM_OUT) ||
            param.mode == CORBA(PARAM_INOUT))
       {
          try
          {
             any2lua(L, param.type, arguments->item(i)->value());
             nresults++;
          }
          catch (Err_Param& ex)
          {
             ex.index = i+1;
             throw;
          }
      }
    }
    return nresults;
}

/*
** Unmarhalled parameters are returned in the Lua stack
**   (the number of parameters pushed in the stack is returned) 
*/
unsigned int MethodDescr::unmarshal_params(lua_State *L, 
					   CORBA::ServerRequest_ptr request,
					   CORBA::NVList_ptr& arguments)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
   unsigned int nparams = 0;

   loCfg->orb->create_list( 0, arguments );
   for(unsigned long i=0; i < this->orb_descr->parameters.length(); i++)
   {
      CORBA(Flags) flags;
      CORBA::ParameterDescription& param = this->orb_descr->parameters[i];

      switch( param.mode )
      {
         case CORBA(PARAM_IN):
            flags = CORBA(ARG_IN);
         break;

         case CORBA(PARAM_OUT):
            flags = CORBA(ARG_OUT);
         break;

         case CORBA(PARAM_INOUT):
            flags = CORBA(ARG_INOUT);
         break;
      }
      CORBA::NamedValue_ptr nv = arguments->add_item(param.name,flags);
#ifdef ORB_TAO
      nv->value()->type(param.type);
#else
      nv->value()->replace(param.type,NULL);
#endif
   }

   #ifdef CORBA_2_3
   request->arguments( arguments );
   #else
   request->params( arguments );
   #endif

   CORBA::ULong nParams = arguments->count();
   for(i=0; i<nParams; i++)
   {
      CORBA::ParameterDescription& param = this->orb_descr->parameters[i];
      if (param.mode != CORBA(PARAM_OUT))
      {
         CORBA::NamedValue_ptr nv = arguments->item(i);
         any2lua(L, param.type, nv->value() );
	 nparams++;
      }
   }
   return nparams;
}

/*
** Results to be marshalled are in the Lua Stack
**   ( "nresults" values, beginning at stack position "pos")
*/
void MethodDescr::marshal_result(lua_State *L, int pos, unsigned int nresults,
				 CORBA::ServerRequest_ptr request,
				 CORBA::NVList_ptr arguments)
{
   CORBA::Any* result = NULL;

   // First result is the method result value (as defined by the IDL)
   // The following results are "out" parameters

   if(this->orb_descr->result &&
      this->orb_descr->result->kind() != CORBA(tk_void))
   {
      if (nresults == 0)
         throw Err_NParams(-1, -1);

      try {
         result = lua2any(L, pos, this->orb_descr->result);
         pos++;
	 nresults--;
      }
      catch (Err_Param& ex)
      {
         ex.index = 0;
         throw;
      }
   }

   for (unsigned int i=0;i<this->orb_descr->parameters.length();i++)
   {
      CORBA::ParameterDescription& param = this->orb_descr->parameters[i];
      if ( param.mode == CORBA(PARAM_OUT) || param.mode == CORBA(PARAM_INOUT))
      {
         if (nresults == 0)
            throw Err_NParams(-1, -1);
         try
         {
            CORBA::Any* any_ptr = lua2any(L, pos, param.type);
            *(arguments->item(i)->value()) = *(any_ptr);
            delete any_ptr;
            pos++;
	    nresults--;
         }
         catch (Err_Param& ex)
         {
            ex.index = i+1;
            throw;
         }
      }
   }

   if (result)
   {
#ifdef CORBA_2_3
      request->set_result(*result);
      delete result;
#else
      request->result(result);
#endif
   }
}

//**************************** Class AttribDescr *********************************

/*
** The parameter to be marshalled is in the Lua stack, at position "pos"
*/
void AttribDescr::marshal_params(lua_State *L,int pos, CORBA::NVList_ptr arguments)
{
    CORBA::Any_var anyValue = lua2any(L, pos, this->orb_descr->type.in());
#ifdef ORB_TAO
    arguments->add_value(this->name, anyValue.in(), CORBA(ARG_IN));
#else
    arguments->add_value(this->name, anyValue, CORBA(ARG_IN));
#endif
}

CORBA::Any* AttribDescr::mount_result(CORBA::Request_ptr req)
{
   if(this->orb_descr->type)
   {
      CORBA::NamedValue* result = req->result();
      CORBA::Any*  resultAny = result->value();

   #ifdef CORBA_2_3
      req->set_return_type(this->orb_descr->type);
   #else
      resultAny->replace(this->orb_descr->type.in(), &result);
   #endif

      return resultAny;
   }
   else
      return NULL;
}

/*
**  The unmarshalled result is returned at the top of the Lua stack
*/

void AttribDescr::unmarshal_result(lua_State *L, CORBA::Any* resultAny)
{
   if (resultAny)
      any2lua(L, this->orb_descr->type.in(), resultAny);
   else
      lua_pushnil(L);
}

/*
**  The unmarshalled parameter is returned at the top of the Lua stack
*/

void AttribDescr::unmarshal_params(lua_State *L, 
				   CORBA::ServerRequest_ptr request)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);

   CORBA::NVList_ptr nvList;
   loCfg->orb->create_list(0,nvList);
#ifdef ORB_TAO
   nvList->add(CORBA(ARG_IN))->value()->type(this->orb_descr->type);
#else
   nvList->add(CORBA(ARG_IN))->value()->replace(this->orb_descr->type,0);
#endif

   #ifdef CORBA_2_3
   request->arguments(nvList);
   #else
   request->params(nvList);
   #endif

   any2lua(L, this->orb_descr->type.in(), nvList->item(0)->value());
}

/*
** The result to be marshalled is in the Lua stack, at position "pos"
*/
void AttribDescr::marshal_result(lua_State *L, int pos, 
				 CORBA::ServerRequest_ptr request)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
   CORBA::NVList_ptr arguments;

   loCfg->orb->create_list( 0, arguments );

#ifdef CORBA_2_3
   request->arguments(arguments);
#else
   request->params(arguments);
#endif

#ifdef CORBA_2_3
   CORBA::Any_var result;
#else
   CORBA::Any* result;
#endif

   result = lua2any(L, pos,  this->orb_descr->type);

#ifdef CORBA_2_3
# ifdef ORB_TAO
   request->set_result(result.in());
# else
   request->set_result(result);
# endif
#else
   request->result(result);
#endif
}
