//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_class.h,v 1.13.2.1 2001/08/07 13:33:02 moura Exp $
//
//*****************************************************************************


#ifndef LO_CLASS_H
#define LO_CLASS_H

#include <stdlib.h>

extern "C"{
#include <lua.h>
}

class MethodDescr
{
public:
   char* name;
   CORBA::OperationDescription* orb_descr;

   virtual ~MethodDescr()
   {
      if (this->name)
         CORBA::string_free(this->name);
      if (this->orb_descr)
         delete this->orb_descr;
   }

   int is_a(char* n)
   { return (strcmp(n, this->name) == 0);}

   CORBA::Any* mount_result(CORBA::Request_ptr req);
   void marshal_params(lua_State *L, int pos, unsigned int nparams,
                       CORBA::NVList_ptr arguments);
   unsigned int  unmarshal_result(lua_State *L, CORBA::Any* resultAny,
				  CORBA::NVList_ptr arguments);

   unsigned int unmarshal_params(lua_State *L, CORBA::ServerRequest_ptr request,
                                 CORBA::NVList_ptr& arguments);
   void marshal_result(lua_State *L, int pos, unsigned int nresults,
		       CORBA::ServerRequest_ptr request,
		       CORBA::NVList_ptr arguments);
};

class AttribDescr
{
public:
   char* name;
   CORBA::AttributeDescription* orb_descr;

   virtual ~AttribDescr()
   {
      if (this->name)
         CORBA::string_free(this->name);
      if (this->orb_descr)
         delete this->orb_descr;
   }

   int is_a(char* n)
   { return (strcmp(n, this->name) == 0);}

   void marshal_params(lua_State *L, int pos, CORBA::NVList_ptr arguments);
   CORBA::Any* mount_result(CORBA::Request_ptr req);
   void unmarshal_result(lua_State *L, CORBA::Any* resultAny);

   void unmarshal_params(lua_State *L, CORBA::ServerRequest_ptr request);
   void marshal_result(lua_State *L, int pos, CORBA::ServerRequest_ptr request);
};

template<class T> class Cache
{
   class Elem
   {
   public:
      T val;
      Elem* next;
      Elem* prev;
      Elem()
      { this->val =0; this->next = 0; this->prev = 0;}
      ~Elem()
       { if (val) delete val; }
   };
   Elem* first;

   void remove (Elem* e)
   {
      if (e->next)
         e->next->prev= e->prev;

      if (e->prev)
         e->prev->next= e->next;

      if (e == this->first)
         this->first= e->next;
   
      e->next= 0;
      e->prev= 0;
      delete e;
   }

public:

   Cache()
   {
      this->first = 0;
   }

   ~Cache()
   {
      while (this->first)
         this->remove(this->first);
   }

   void insert(T val)
   {
      Elem* elem = new Elem;
      elem->val = val;
      elem->next = this->first;
      if(this->first)
         this->first->prev = elem;
      this->first = elem;
   }

   T get(char* name)
   {
      Elem* current = this->first;
      while (current)
      {
         if (current->val->is_a(name))
            return current->val;
         current = current->next;
      }
      return 0;
   }

   void remove(T val)
   {
      Elem* current = this->first;
      while (current)
      {
         if (val == current->val)
         {
            this->remove(current);
            return;
         }
         current = current->next;
      }
   }
};

typedef Cache<MethodDescr*> MethodCache;
typedef Cache<AttribDescr*> AttribCache;

class LuaOrbClass
{
   void init(lua_State *L);
   MethodDescr* add_opdef(char*,CORBA::OperationDef_ptr);
   AttribDescr* add_attdef(char*,CORBA::AttributeDef_ptr);

   char* _repository_id;
   char* _interface_name;
   CORBA::InterfaceDef_ptr _interfacedef;
   MethodCache* method_cache;
   AttribCache* attr_cache;
   LuaOrbClass* next;

   // Registry data
   static const char* reg_CLASSES;
   static LuaOrbClass* getFirstClass(lua_State *L);
   static void setFirstClass(lua_State *L, LuaOrbClass* loclass);

public:

   LuaOrbClass()
   {
      this->_repository_id = NULL;
      this->_interface_name = NULL;
      this->_interfacedef = CORBA::InterfaceDef::_nil();
      this->method_cache = new MethodCache;
      this->attr_cache = new AttribCache;
      this->next = NULL;
   }

   virtual ~LuaOrbClass()
   {
      if (this->method_cache)
         delete this->method_cache;
      if (this->attr_cache)
         delete this->attr_cache;
      if (this->_repository_id)
         CORBA::string_free(this->_repository_id);
      if (this->_interface_name)
         CORBA::string_free(this->_interface_name);
      if (!CORBA::is_nil(this->_interfacedef))
         CORBA_release(this->_interfacedef);
   }

   static void close(lua_State *L);
   static LuaOrbClass* get_class_by_id(lua_State *L, const char* id);
   static LuaOrbClass* get_class_by_name(lua_State *L, const char* iname);
   static LuaOrbClass* get_class_by_obj(lua_State *L, CORBA::Object_ptr obj);
   static LuaOrbClass* get_class_by_interfacedef(lua_State *L, 
                                                 CORBA::InterfaceDef_ptr idef);

   MethodDescr* get_opdef(lua_State *L, char* name);
   AttribDescr* get_attdef(char* name);

   char* interface_name() {
      return this->_interface_name;
   }
   void interface_name(const char* iname) {
      if (this->_interface_name)
         CORBA::string_free(this->_interface_name);
      if (iname && iname[0] == ':' && iname[1] == ':')
         iname = &iname[2];
      this->_interface_name = CORBA::string_dup(iname);
   }

   char* repository_id() {
      return this->_repository_id;
   }
   void repository_id(const char* id) {
      if (this->_repository_id)
         CORBA::string_free(this->_repository_id);
      this->_repository_id = CORBA::string_dup(id);
   }

   CORBA::InterfaceDef_ptr get_interfacedef();
};
#endif
