//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_dsi.cpp,v 1.36.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************

#include <assert.h>
#include <stdio.h>

extern "C"
{
#include <lua.h>
#include <lauxlib.h>
}

#include "orb_inc.h"
#include "lo_dsi.h"
#include "lo_proxy.h"
#include "lo_cfg.h"
#include "lo_map.h"
#include "lo_except.h"
#include "lo_util.h"
#include "luawrap.h"

#include "lo_out.h"

#ifdef TEST_PERFORMANCE
#include <time.h>

#define _BEGIN_TEST_PERFORMANCE(s)      \
   cout << "#DSI Calling " << s << endl;
   clock_t _T1 = clock();               \
   clock_t _T2 = _T1;                   \

#define _MARK_TIME     \
   _T2 = clock();

#define _PRINT_TIME(s)   \
   cout << s << (clock()-_T2)/CLOCKS_PER_SEC << endl;

#define _END_TEST_PERFORMANCE(s)  \
   cout << "   Total: " << (clock()-_T1)/CLOCKS_PER_SEC << endl;  \
   cout << "#Finish " << s << endl;

#else
#define _BEGIN_TEST_PERFORMANCE(s) 
#define _MARK_TIME
#define _PRINT_TIME(msg)
#define _END_TEST_PERFORMANCE(s)
#endif

#ifdef ORB_ORBACUS
OB_MAKE_NARROW_IMPL_1( LuaDSI, CORBA::DynamicImplementation )
#endif

/*******************************************************************/
/*   Prototypes for functions registered in Lua			   */
/*******************************************************************/

static int _lo_createservant(lua_State *L);

static int index_fb(lua_State *L);
static int settable_fb(lua_State *L);
static int gc_fb(lua_State *L);

static int _get_ior(lua_State *L);
static int _get_interface_name(lua_State *L);
static int _get_repository_id(lua_State *L);
static int _get_interface(lua_State *L);
static int _get_corba_ref(lua_State *L);
static int _get_implementation(lua_State *L);
static int _set_implementation(lua_State *L);


/*******************************************************************/
/*   LuaWrapper data 					           */
/*******************************************************************/

const char* LuaDSI::wrap_LuaDSI = "LuaOrb_DSIWrapper";

static struct luaL_reg functions_tb []= {{"lo_createservant",(lua_CFunction)_lo_createservant}};

static struct luaL_reg tagmethods_tb[]= {
   {"index",(lua_CFunction)index_fb},
   {"settable",(lua_CFunction)settable_fb}
};

static struct luaL_reg ud_tagmethods_tb[]= {{"gc",(lua_CFunction)gc_fb}};

static struct luaL_reg methods_tb[]= {
   {"_get_ior",(lua_CFunction)_get_ior},
   {"_get_interface_name",(lua_CFunction)_get_interface_name},
   {"_get_id",(lua_CFunction)_get_repository_id},
   {"_get_interface",(lua_CFunction)_get_interface},
   {"_get_corba_ref",(lua_CFunction)_get_corba_ref},
   {"_get_implementation",(lua_CFunction)_get_implementation},
   {"_set_implementation",(lua_CFunction)_set_implementation}
};



/**************************************************************************/
/*                                                                        */
/*  Class LuaDSI                                                          */
/*                                                                        */
/**************************************************************************/

/*
** Registry data
*/

const char* LuaDSI::reg_EXCEPTION_TAG = "LuaOrb_EXCEPTION_TAG";
const char* LuaDSI::reg_SERVERLIST = "LuaOrb_ServerList";

/*
** Methods
*/

LuaDSI* LuaDSI::CreateImplicitServant(lua_State *L, int pos, char* rep_id)
{
   LuaDSI* servant = NULL;

   LuaOrbCfg::getNewServantTagMethod(L, lua_tag(L, pos));
   if (lua_isfunction(L, -1))  // tag method at the top of the stack
   {
      LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
      LuaOrbClass* lo_class = LuaOrbClass::get_class_by_id(L, rep_id);
      lua_pushvalue(L, pos);
      lua_pushstring(L, rep_id);
      if (lo_class && lo_class->interface_name())
         lua_pushstring(L, lo_class->interface_name());
      else
         lua_pushnil(L);
      lua_call(L,3,1);		// 3 args and 1 result
      servant = (LuaDSI*) Wrapper->from_lua(L, lua_gettop(L));
      lua_pop(L, 1);	
   }
   else
      servant = LuaDSI::Create(L, pos, rep_id);

   return servant;
}

LuaDSI* LuaDSI::Create(lua_State *L, int pos, const char* InterfaceName, char* name)
{
   LuaOrbClass* lo_class = NULL;

   if (LuaOrbUtil::isRepositoryId(InterfaceName))
      lo_class = LuaOrbClass::get_class_by_id(L, InterfaceName);
   else
      lo_class = LuaOrbClass::get_class_by_name(L, InterfaceName);

   if (CORBA::is_nil(lo_class->get_interfacedef()))
   {
      char msg[100];
      sprintf(msg,"The Interface %s does not exist",InterfaceName);
      throw Err_General(msg);;
   }

   // Create the DSI object
   LuaDSI* server =  new LuaDSI(L, pos, lo_class, name);
   return server;
}

LuaDSI* LuaDSI::Create(lua_State *L, int pos, CORBA::InterfaceDef_ptr intf, char* name)
{
   LuaOrbClass* lo_class = LuaOrbClass::get_class_by_interfacedef(L, intf);

   // Create the DSI object
   LuaDSI* server =  new LuaDSI(L, pos, lo_class, name);
   return server;
}

LuaDSI::LuaDSI( lua_State *L, int pos,  LuaOrbClass* lo_class, char* name)
#ifdef ORB_VISIBROKER
: CORBA_DynamicImplementation((const char*)lo_class->interface_name(),name,(const char*)lo_class->repository_id())
#endif
{

   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);

   // Lua State and the "Exception TAG" are recorded in the LuaDSI object 
   //   because the "invoke" operation needs them

   myluaState = L;
   lua_getregistry(L);
   lua_pushstring(L, LuaDSI::reg_EXCEPTION_TAG);
   lua_gettable(L, -2);
   EXCEPTION_TAG = (int) lua_tonumber(L, -1);
   lua_pop(L, 2);

   myclass = lo_class;
   tableLuaRef = -1;
   this->SetLuaObject(L, pos);

#ifdef CORBA_2_3
   if (name)
   {
      if (loCfg->persistent_poa)
         this->poa = loCfg->persistent_poa;
      else
         this->poa = loCfg->poa;

      PortableServer::ObjectId_var oid = PortableServer::string_to_ObjectId(name);
      this->poa->activate_object_with_id(oid,this);

#ifdef ORB_ORBACUS
      CORBA::Object_var obj_ref = loCfg->orb->resolve_initial_references("BootManager");
      OB::BootManager_var bootManager = OB::BootManager::_narrow(obj_ref);
      CORBA::Object_var oref = this->poa->servant_to_reference(this);
      bootManager->add_binding(oid,oref);
#endif
   }
   else
   {
      this->poa = loCfg->poa;
      PortableServer::ObjectId_var oid = this->poa->activate_object(this);
   }
#elif defined ORB_ORBACUS
   if (name)
      loCfg->orb->connect(this,name);
   else
      loCfg->orb->connect(this);

#elif defined ORB_MICO
   loCfg->boa->obj_is_ready(this,0);
#else
   loCfg->boa->obj_is_ready(this);
#endif

   LuaDSI::insertServer(L, this);
}

LuaDSI::~LuaDSI()
{
   lua_State *L = myluaState;
 
   lua_unref( L, tableLuaRef );
   if (this->nextServer || this->prevServer)
      LuaDSI::removeServer(L, this);

   #ifdef CORBA_2_3
//   PortableServer::ObjectId_var oid = this->poa->servant_to_id(this);
//   this->poa->deactivate_object(oid);
   #else
   (LuaOrbCfg::getConfiguration(L))->boa->deactivate_obj(this);
   #endif
}

void LuaDSI::GetLuaObject(lua_State *L)
{
   lua_getref(L, this->tableLuaRef);
}

void LuaDSI::SetLuaObject(lua_State *L, int pos)
{
   /* Discard (unref) previous Lua Object */
   if (tableLuaRef != -1)
   {
      lua_getref(L, this->tableLuaRef);
      lua_pushstring(L, "_last_lo_servant");
      lua_pushnil(L);
      lua_rawset(L, -3);
      lua_pop(L, 1);

      lua_unref(L, this->tableLuaRef);
      this->tableLuaRef = -1;
#ifdef CORBA_2_3
      this->_remove_ref();
#else
      CORBA_release(this);
#endif
   }
   if(!LuaOrbUtil::isIndexable(L, pos))
      return;

   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);

   lua_pushvalue(L, pos);
   this->tableLuaRef = lua_ref(L, 1);

   lua_pushstring(L, "_last_lo_servant");
   Wrapper->push(L, this);
   lua_rawset(L, pos);
#ifdef CORBA_2_3
   this->_add_ref();
#else
   LuaDSI::_duplicate(this);
#endif
}

#ifdef ORB_TAO
CORBA::Boolean LuaDSI::_is_a(const char *name,CORBA::Environment&)
#else
CORBA::Boolean LuaDSI::_is_a(const char *name)
#endif
{
   CORBA::InterfaceDef_ptr intfDef = this->_get_interface();
   CORBA::Boolean result = intfDef->is_a( name );
   return result;
}

#ifdef ORB_TAO
CORBA::InterfaceDef_ptr LuaDSI::_get_interface(CORBA::Environment&)
#else
CORBA::InterfaceDef_ptr LuaDSI::_get_interface()
#endif
{
   return CORBA::InterfaceDef::_duplicate(this->myclass->get_interfacedef());
}

#ifdef CORBA_2_3
#ifdef ORB_TAO
CORBA::RepositoryId LuaDSI::_primary_interface(const PortableServer::ObjectId& oid, PortableServer::POA_ptr _poa, CORBA::Environment&)
#else
CORBA::RepositoryId LuaDSI::_primary_interface(const PortableServer::ObjectId& oid, PortableServer::POA_ptr _poa)
#endif
{
   return CORBA::string_dup(this->myclass->repository_id());
}

#ifdef ORB_TAO
PortableServer::POA_ptr LuaDSI::_default_POA(CORBA::Environment&)
#else
PortableServer::POA_ptr LuaDSI::_default_POA()
#endif
{
   return PortableServer::POA::_duplicate(this->poa);
}

#endif


#ifdef ORB_TAO
void LuaDSI::invoke(CORBA::ServerRequest_ptr request, CORBA::Environment&)
#else
void LuaDSI::invoke(CORBA::ServerRequest_ptr request)
#endif
{
   lua_State *L = myluaState;
   int top = lua_gettop(L);

   const char* op_name = NULL;
   try
   {
#ifdef CORBA_2_3
      op_name = request->operation();
#else
      op_name = request->op_name();
#endif

      _BEGIN_TEST_PERFORMANCE(op_name)

      MethodDescr* op_descr = this->myclass->get_opdef(L, (char*)op_name);
      if (op_descr)
      {
         VERBOSE_OUT2(L, "Invoke Operation: ", op_name)

         // Get the Lua function that implements the operation
         lua_getref(L, tableLuaRef);
         lua_pushstring(L, op_name);
	 lua_gettable(L, -2);		// Lua function
	 lua_insert(L, -2);		//  self is the  1st parameter

	 // the unmarshalled parameters are pushed onto the stack 
         _MARK_TIME
         CORBA::NVList_ptr nvList;
         unsigned int nparams = op_descr->unmarshal_params(L, request, nvList);
         _PRINT_TIME("   Unmarshalling Parameters: ")

         if (lua_call(L, nparams+1, LUA_MULTRET))
         {
#if defined(ORB_ORBACUS) && !defined(CORBA_2_3)
            CORBA::Any *any = new CORBA::Any;
            *any <<= new CORBA_UNKNOWN;
            request->exception( any );
	    lua_settop(L, top);
            return;
#else
            throw CORBA::UNKNOWN();
#endif
         }
	 // Function and arguments (including "self") were popped
	 //   Lua stack now contains the results

         if (lua_tag(L, top+1) == this->EXCEPTION_TAG)
            this->RaiseException(L, request,op_descr);
         else
         {
            _MARK_TIME
 	    unsigned int nresults = lua_gettop(L) - top;  // # results
            op_descr->marshal_result(L, top + 1, nresults, request, nvList);
            _PRINT_TIME("   Marshalling Results: ")
         }
      }
      else
      {
         // Remove _get_ and _set_ and try an attribute
         char* name = (char*)op_name + 5;
         AttribDescr* attr_descr = this->myclass->get_attdef(name);
         if (attr_descr)
         {
            lua_getref(L, tableLuaRef);		// push Lua table
            lua_pushstring(L, name);		//  and index (attribute name)
            // Set ?
            if( strncmp( "_set_", op_name, 5 )==0 )
            {
               VERBOSE_OUT2(L, "Setting Attribute: ", name) 
               attr_descr->unmarshal_params(L, request);
               lua_settable(L, -3);
            }
            else // Get 
            {
               VERBOSE_OUT2(L, "Getting Attribute: ", name) 
               lua_gettable(L, -2);
               attr_descr->marshal_result(L, lua_gettop(L), request);
            }
         }
         else
         {
            VERBOSE_OUT2(L, "Error calling: ", op_name)

#if defined(ORB_ORBACUS) && !defined(CORBA_2_3)
               CORBA::Any *any = new CORBA::Any;
               *any <<= new CORBA_NO_IMPLEMENT;
               request->exception( any );
#else
               throw CORBA::NO_IMPLEMENT();
#endif
         }
      }
      _END_TEST_PERFORMANCE(op_name)
   }
   catch( CORBAEXP(BAD_PARAM) )
   {
      LO_OUT(L) << "Exception Bad Param" << endl;
      throw;
   }
   catch( CORBAEXP(MARSHAL) )
   {
      LO_OUT(L) << "Exception Marshal" << endl;
      throw;
   }
   catch( CORBAEXP(BAD_TYPECODE) )
   {
      LO_OUT(L) << "Exception bad typecode" << endl;
      throw;
   }
   catch( CORBAEXP(BAD_OPERATION) )
   {
      LO_OUT(L) << "Exception bad operation" << endl;
      throw;
   }
   catch( CORBA::SystemException& )
   {
      LO_OUT(L) << "System Exception ERROR!" << endl;   
      throw;
   }
   catch(Err_NParams&)
   {
      LO_OUT(L) << "Exception in Operation " << op_name << ": Invalid number of results." << endl;
      throw CORBAEXP(UNKNOWN)();
   }
   catch(Err_Param& excep)
   {
      if (excep.index==0)
         LO_OUT(L) << "Exception in Operation " << op_name << ": Invalid result type." << endl;
      else
         LO_OUT(L) << "Exception in Operation " << op_name << ": Invalid out parameter (" << excep.index << ")." << endl;
      throw CORBAEXP(UNKNOWN)();
   }
   catch (Err_General& excep)
   {
      LO_OUT(L) << excep.msg << endl;
#if defined(ORB_ORBACUS) && !defined(CORBA_2_3)
      CORBA::Any *any = new CORBA::Any;
      *any <<= new CORBA_UNKNOWN;
      request->exception( any );
      lua_settop(L, top);
      return;
#else
      throw CORBA::UNKNOWN();
#endif
   }
   catch( ... )
   {
      LO_OUT(L) << "Exception ERROR!" << endl;
      throw;
   }
   lua_settop(L, top);
}

/*
** The Lua object that describes the exception is at the top of the stack
*/
void LuaDSI::RaiseException(lua_State *L, CORBA::ServerRequest_ptr request,
			    MethodDescr* descr)
{
   if (descr->orb_descr->exceptions.length() >= 1)
   {
      unsigned int i;
      lua_pushnumber(L, 1.0);
      lua_gettable(L, -2);
      const char* ex_name = lua_tostring(L, -1);
      lua_pop(L, 1);

      if (ex_name)
      {
         for (i = 0; i < descr->orb_descr->exceptions.length(); i++)
         {
            if (strcmp(ex_name,descr->orb_descr->exceptions[i].type->name())==0)
            {
               CORBA::Any* anyValue = lua2any(L, lua_gettop(L), 
					      descr->orb_descr->exceptions[i].type);

               #ifdef CORBA_2_3
               const CORBA::Any any = *anyValue;
               request->set_exception(any);
               delete anyValue;
               #else
               request->exception(anyValue);
               #endif      

               return;
            }
         }
      }
   }
}

void LuaDSI::open(lua_State *L)
{
   static char init_code[] = "function LuaOrbException(t)\n"
                                   "   settag(t,LUAORB_EXCEPTION_TAG)\n"
                                   "   return t\n"
                                   "end";
   luaL_openlib(L, functions_tb,sizeof(functions_tb)/sizeof(struct luaL_reg));

   LuaWrapper* Wrapper = LuaWrapper::createWrapper(L, wrap_LuaDSI);
   Wrapper->reg_tagmethods(L, tagmethods_tb,
                          sizeof(tagmethods_tb)/sizeof(struct luaL_reg));
   Wrapper->reg_ud_tagmethods(L, ud_tagmethods_tb,
                            sizeof(ud_tagmethods_tb)/sizeof(struct luaL_reg));
   Wrapper->reg_methods(methods_tb,
                          sizeof(methods_tb)/sizeof(struct luaL_reg));

   int EXC_TAG = lua_newtag(L);
   lua_pushnumber(L, EXC_TAG);
   lua_setglobal(L, "LUAORB_EXCEPTION_TAG");

   // Store "Exception TAG" in Lua Registry
   lua_getregistry(L);
   lua_pushstring(L, LuaDSI::reg_EXCEPTION_TAG);
   lua_pushnumber(L, EXC_TAG);
   lua_settable(L, -3);
   lua_pop(L, 1);

   lua_dostring(L, init_code);
}

void LuaDSI::close(lua_State *L)
{
   LuaWrapper::deleteWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI::destroyServerList(L);
}

LuaDSI* LuaDSI::getFirstServer(lua_State *L)
{
   LuaDSI* firstServer = NULL;

   lua_getregistry(L);
   lua_pushstring(L, LuaDSI::reg_SERVERLIST);
   lua_gettable(L, -2);
   if (!lua_isnil(L, -1))
      firstServer = (LuaDSI*) lua_touserdata(L, -1);
   lua_pop(L, 2);
   return firstServer;
}

void LuaDSI::setFirstServer(lua_State *L, LuaDSI* server)
{
   lua_getregistry(L);
   lua_pushstring(L, LuaDSI::reg_SERVERLIST);

   if (server == NULL)
      lua_pushnil(L);
   else
      lua_pushuserdata(L, server);
   lua_settable(L, -3);
   lua_pop(L, 1);
}

void LuaDSI::destroyServerList(lua_State *L)
{
   LuaDSI* server = LuaDSI::getFirstServer(L);

   while(server)
   {
#ifdef CORBA_2_3
      PortableServer::ObjectId_var oid = server->poa->servant_to_id(server);
      server->poa->deactivate_object(oid);
      server->_remove_ref();
#else
      (LuaOrbCfg::getConfiguration(L))->boa->deactivate_obj(server);
      CORBA_release(server);
#endif
      server = server->nextServer;
   }
   LuaDSI::setFirstServer(L, NULL);
}

void LuaDSI::insertServer(lua_State *L, LuaDSI* server)
{
   LuaDSI* firstServer = LuaDSI::getFirstServer(L);
   server->nextServer = firstServer;
   server->prevServer = 0;
   if (firstServer != NULL)
      firstServer->prevServer = server;
   else
      LuaDSI::setFirstServer(L, server);
}

void LuaDSI::removeServer(lua_State *L, LuaDSI* server)
{
   LuaDSI* firstServer = LuaDSI::getFirstServer(L);

   if (server->nextServer)
      server->nextServer->prevServer = server->prevServer;

   if (server == firstServer)
      LuaDSI::setFirstServer(L, server->nextServer);
   else
      server->prevServer->nextServer = server->nextServer;
   
   server->nextServer = 0;
   server->prevServer = 0;
}

/***************************************************************************/
/*   Create a LuaDSI object (explicit dynamic server creation)		   */
/***************************************************************************/

static int _lo_createservant(lua_State *L)
{
   const char* interface_name = luaL_check_string(L, 2);
   char* name = NULL;

   luaL_arg_check(L, LuaOrbUtil::isIndexable(L, 1), 1,
                  "table or indexable userdata expected");

   if (lua_isstring(L, 3))
      name = (char *)lua_tostring(L, 3);

   try {
      LuaDSI* dserver = LuaDSI::Create(L, 1, interface_name, name);

      #ifdef CORBA_2_3
      dserver->_add_ref();
      #else
      LuaDSI::_duplicate( dserver );
      #endif

      LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
      Wrapper->push(L, dserver);
   }
   catch (Err_General& excep)
   {
      lua_error(L, excep.msg);
   }
   #ifdef CORBA_2_3
   catch(const CORBA::Exception & excep )
   {
      LO_OUT(L) << excep <<endl;
      lua_error(L, "CORBA exception ERROR");
   }
   #endif

   return 1;
}

/***************************************************************************/
/*  Tag-Methods for LuaDSI objects                                         */
/***************************************************************************/

static int gc_fb(lua_State *L)
{
   LuaDSI* dserver = (LuaDSI*) lua_touserdata(L, 1);

#ifdef CORBA_2_3
   dserver->_remove_ref();
#else
   CORBA_release(dserver);
#endif
   return 0;
}

static int index_fb(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);

   dserver->GetLuaObject(L);	// push the Lua object (server implementation)
   lua_pushvalue(L, -2);	//   and push the index
   lua_gettable(L, -2);
   lua_remove(L, -2);		// remove Lua object from stack

   return 1;			// requested value at the top of the stack
}

static int settable_fb(lua_State *L)
{
   int top = lua_gettop(L);

   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->from_lua(L, 1);

   if (lua_isstring(L, 2))
   {
      const char* field = lua_tostring(L, 2);
      int n = sizeof(methods_tb)/sizeof(struct luaL_reg);
      for (int i=0; i < n; i ++)
      {
          if ( strcmp(methods_tb[i].name, field) == 0 )
          {
	     lua_pushvalue(L, 2);
	     lua_pushvalue(L, 3);
             lua_rawset(L, 1);
          } 
      }
   }
   dserver->GetLuaObject(L);	// Lua object (server implementation)
   lua_pushvalue(L, 2);		//  index
   lua_pushvalue(L, 3);		//  value
   lua_settable(L, -3);

   lua_settop(L, top);
   return 0;
}

/***************************************************************************/
/*  Pre-defined operations for LuaDSI dynamic servers			   */
/***************************************************************************/

static int _get_ior(lua_State *L)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);

   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);

   #ifdef CORBA_2_3
   PortableServer::POA_var poa = dserver->_default_POA();
   CORBA::Object_var obj = poa->servant_to_reference(dserver);
   char* ior = loCfg->orb->object_to_string(obj);
   #else
   char* ior = loCfg->orb->object_to_string(dserver);
   #endif

   lua_pushstring(L, ior);
   CORBA::string_free(ior);
   return 1;
}

static int _get_repository_id(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*)Wrapper->check_tag(L, 1);

   if (dserver->myclass->repository_id())
      lua_pushstring(L, dserver->myclass->repository_id());
   else
      lua_pushnil(L);
   return 1;
}

static int _get_interface_name(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);

   if (dserver->myclass->interface_name())
      lua_pushstring(L, dserver->myclass->interface_name());
   else
      lua_pushnil(L);
   return 1;
}

static int _get_interface(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);

   LuaProxy* newproxy = new LuaProxy(L, "CORBA::InterfaceDef",
                                        dserver->myclass->get_interfacedef());
   if (newproxy)
   {
      newproxy->locked = 1;
      LuaProxy::push(L, newproxy);
   }
   else
      lua_pushnil(L);
   return 1;
}

static int _get_corba_ref(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);

   lua_pushuserdata(L, dserver);
   return 1;
}

static int _get_implementation(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);
   dserver->GetLuaObject(L);
   return 1;
}

static int _set_implementation(lua_State *L)
{
   LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
   LuaDSI* dserver = (LuaDSI*) Wrapper->check_tag(L, 1);
   luaL_checktype(L, 2, LUA_TTABLE);
   dserver->SetLuaObject(L, 2);
   return 0;
}
