//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_proxy.cpp,v 1.56.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <orb_inc.h>

extern "C"
{
#include <lua.h>
#include <lauxlib.h>
}

#include "lo_map.h"
#include "lo_proxy.h"
#include "lo_cfg.h"
#include "lo_except.h"
#include "lo_util.h"

#include "lo_out.h"

#ifdef TEST_PERFORMANCE
#include <time.h>

#define _BEGIN_TEST_PERFORMANCE(s)      \
   cout  << "#DII Calling " << s << endl;
   clock_t _T1 = clock();               \
   clock_t _T2 = _T1;                   \

#define _MARK_TIME     \
   _T2 = clock();

#define _PRINT_TIME(s)   \
   cout << s << (clock()-_T2)/CLOCKS_PER_SEC << endl;

#define _END_TEST_PERFORMANCE(s)  \
   cout << "   Total: " << (clock()-_T1)/CLOCKS_PER_SEC << endl;  \
   cout << "#Finish " << s << endl;

#else
#define _BEGIN_TEST_PERFORMANCE(s) 
#define _MARK_TIME
#define _PRINT_TIME(msg)
#define _END_TEST_PERFORMANCE(s)
#endif

static int _get_ior(lua_State *L);
static int _get_interface_name(lua_State *L);
static int _get_interface_def(lua_State *L);
static int _get_repository_id(lua_State *L);
static int _get_corba_ref(lua_State *L);
static int _non_existent(lua_State *L);


//**************************** Class LuaProxy *********************************

const char* LuaProxy::reg_TAG   = "LuaOrb_PROXY_TAG";
const char* LuaProxy::reg_UDTAG = "LuaOrb_PROXY_UDTAG";

const char* LuaProxy::wrap_DEFCLOSURE = "LuaOrb_DEFWrapper";
 
int LuaProxy::isProxy(lua_State *L, int pos)
{
   lua_getregistry(L);
   lua_pushstring(L, LuaProxy::reg_TAG);
   lua_gettable(L, -2);
   int res = (lua_tag(L, pos) == (int) lua_tonumber(L,-1));
   lua_pop(L, 2);
   return res;
}

void LuaProxy::push(lua_State *L, LuaProxy* proxy)
{
   proxy->GetLuaThis(L);
}

LuaProxy* LuaProxy::check_tag(lua_State *L, int narg, int tag)
{
   luaL_arg_check(L, (lua_tag(L,narg) == tag),narg,"it isn't a Corba Object");
   LuaProxy* proxy = LuaProxy::from_lua(L, narg);
   return proxy;
}

LuaProxy* LuaProxy::from_lua(lua_State *L, int pos)
{
   if (lua_istable(L,pos))
   {
      lua_pushstring(L, "_CORBA_OBJ_REF_");
      lua_gettable(L, pos);
      LuaProxy* proxy = (LuaProxy*) lua_touserdata(L, -1);
      lua_pop(L, 1);
      return proxy;
   }
   else
      return NULL;
}

LuaProxy::LuaProxy(lua_State *L, const char* interfacename, CORBA::Object_ptr orb_obj)
: orb_object(orb_obj)
{
   this->locked = 0;

   if (interfacename == NULL)
      this->myclass = LuaOrbClass::get_class_by_obj(L, this->orb_object);
   else if (LuaOrbUtil::isRepositoryId(interfacename))
      this->myclass = LuaOrbClass::get_class_by_id(L, interfacename);
   else
      this->myclass = LuaOrbClass::get_class_by_name(L, interfacename);

   this->CreateLuaThis(L);
}

LuaProxy::~LuaProxy()
{
   CORBA_release(this->orb_object);
}

static struct luaL_reg proxy_methods_tb[]= {
   {"_get_id", (lua_CFunction)_get_repository_id},
   {"_get_interface_name", (lua_CFunction)_get_interface_name},
   {"_get_interface", (lua_CFunction)_get_interface_def},
   {"_get_corba_ref", (lua_CFunction)_get_corba_ref},
   {"_non_existent", (lua_CFunction)_non_existent},
   {"_get_ior", (lua_CFunction)_get_ior},
};

void LuaProxy::CreateLuaThis(lua_State *L)
{
   lua_getregistry(L);
   lua_pushstring(L, LuaProxy::reg_TAG);
   lua_gettable(L, -2);
   int TAG = (int) lua_tonumber(L, -1);
   lua_pushstring(L, LuaProxy::reg_UDTAG);
   int UDTAG = (int) lua_tonumber(L, -1);
   lua_pop(L, 3);

   lua_newtable(L);

   lua_pushstring(L, "_CORBA_OBJ_REF_");
   lua_pushusertag(L, this, UDTAG);
   lua_rawset(L, -3);

   for (unsigned int i=0; i < sizeof(proxy_methods_tb)/sizeof(proxy_methods_tb[0]) ; i++)
   {
       lua_pushstring(L, proxy_methods_tb[i].name);
       lua_pushnumber(L, TAG);
       lua_pushcclosure(L, proxy_methods_tb[i].func,1);
       lua_rawset(L, -3);
   }

   lua_settag(L, TAG);
   this->lua_this = lua_ref(L, 0);
}

void LuaProxy::GetLuaThis(lua_State *L)
{
   lua_getref(L, this->lua_this);
}

void LuaProxy::SetLuaThis(lua_State *L, int pos)
{
   lua_unref(L, this->lua_this);
   lua_pushvalue(L, pos);
   this->lua_this = lua_ref(L, 0);
}

void LuaProxy::DiscardLuaThis(lua_State *L)
{
   lua_unref(L, this->lua_this);
}

char* LuaProxy::parsename(char* full_name, InvokeMode* mode)
{
   static char prefix [] = "deferred_";
   char* name = full_name;

   if (mode!=NULL)
   {
      unsigned int n = strlen(prefix);

      *mode = NORMAL_CALL;

      if (strlen(full_name) > n)
      {
         if (strncmp(prefix,full_name,n) == 0)
         {
            name = & full_name[n];
            *mode = DEFERRED_CALL;
         }
      }
   }

   return name;
}

void LuaProxy::GetLuaExceptionHandler(lua_State *L)
{
   this->GetLuaThis(L);
   lua_pushstring(L, "exception_handler");
   lua_gettable(L, -2);
   lua_remove(L, -2);   // remove Lua proxy table from stack

}

/*
** The return values that will replace the return values of the called
**  methods are returned in the Lua stack
**    (the number of values pushed in the stack is returned)
*/ 
unsigned int LuaProxy::RaiseException(lua_State *L, CORBA::Exception* excep, MethodDescr* descr)
{
   unsigned int nresults = 0;
   int top = lua_gettop(L);

   #ifdef CORBA_2_3
   CORBA::UnknownUserException* u_ex = CORBA::UnknownUserException::_downcast(excep);
   #else
   CORBA::UnknownUserException* u_ex = CORBA::UnknownUserException::_narrow(excep);
   #endif

   this->GetLuaExceptionHandler(L); // exception handler at the top of the stack
   if (u_ex && lua_isfunction(L, -1))
   {
      LuaProxy::push(L, this);  // Lua proxy object
      lua_pushstring(L, descr->name); //the method name
      lua_pushstring(L, (char*)u_ex->exception().type()->name()); //the exception name
      any2lua(L, u_ex->exception().type(),&(u_ex->exception())); //exception's fields

      lua_call(L, 4, LUA_MULTRET);   // exception handler & parameters
                                     //  are popped !

      //  the results from the exception handler are in the Lua stack
      //     find out how many they are
      nresults = lua_gettop(L) - top;
   }
   else if (lua_isfunction(L, -1))
   {
      LuaProxy::push(L, this);  // Lua proxy object
      lua_pushstring(L, descr->name); //the method name
      lua_pushstring(L, (char*)exception_name(*excep)); //the exception name
      lua_call(L, 3, LUA_MULTRET);

      //  the results from the exception handler are in the Lua stack
      //     find out how many they are
      nresults = lua_gettop(L) - top;
   }
   else
   {
      char msg[100];
      sprintf(msg,"LuaOrb Error: The method %s raises an exception %s",
              descr->name, exception_name(*excep));
      throw Err_General(msg);
   }
   return nresults;
}

/*
** The invoked method parameters are in the Lua stack
**   ("nparams" values, beginning at stack position "pos"
**
** The method results are returned in the Lua stack
**  (the number of results pushed in the stack is returned)
*/

unsigned int  LuaProxy::Call(lua_State *L, int pos, unsigned int nparams, 
                             MethodDescr* descr, InvokeMode mode)
{
   unsigned int nresult = 0;

   _BEGIN_TEST_PERFORMANCE(descr->name)
   /* Build the parameter list
      Check the request mode: oneway or not
      Perform the request
      Handle the result, if there is one
      Check if there is an exception
   */
   CORBA::Request_ptr req = this->orb_object->_request(descr->name);
   CORBA::NVList_ptr arguments = req->arguments();

   _MARK_TIME
   descr->marshal_params(L, pos, nparams, arguments); /* build parameter list */
   _PRINT_TIME("   Marshalling Parameters: ")

   if (descr->orb_descr->exceptions.length() >= 1)
   {
      unsigned int i;
      for (i = 0; i < descr->orb_descr->exceptions.length(); i++)
         req->exceptions()->add( descr->orb_descr->exceptions[i].type);
   }

   if (descr->orb_descr->mode == CORBA(OP_NORMAL))
   {
      CORBA::Any* resultAny = descr->mount_result(req);

      if (mode == NORMAL_CALL)
      {
         req->invoke();   /*method request*/

         if (req->env()->exception() == NULL)
         {
            _MARK_TIME
            //it will push the result for Lua
            nresult = descr->unmarshal_result(L, resultAny, arguments);
            _PRINT_TIME("   Unmarshalling Results: ")
         }
         else
            nresult = this->RaiseException(L, req->env()->exception(),descr);
      }
      else
      {
         req->send_deferred();
         LuaProxy::DeferredClosure* closure = new LuaProxy::DeferredClosure;
         closure->myluaState = L;
         closure->object = this;
         closure->req = req;
         closure->resultAny = resultAny;
         closure->arguments = arguments;
         closure->method_descr = descr;

         LuaWrapper* DefClosureWrapper = 
         LuaWrapper::getWrapper(L, LuaProxy::wrap_DEFCLOSURE);
         DefClosureWrapper->push(L, closure);

         return 1; // closure wrapper is returned !
      }
   }
   else
      req->send_oneway(); /*method request -> no return! */

   CORBA_release(req);
   _END_TEST_PERFORMANCE(descr->name)

   return nresult;
}

/**
** SetAttr returns 1 if deferred call (closure wrapper is pushed onto the stack)
*/
int LuaProxy::SetAttr(lua_State *L, int pos, AttribDescr* descr, InvokeMode mode)
{
   char req_name[70];
   sprintf(req_name,"_set_%s",descr->name);

   CORBA::Request_ptr req = this->orb_object->_request(req_name);
   CORBA::NVList_ptr arguments = req->arguments();

   descr->marshal_params(L, pos, arguments); /* build the parameter list */
   if (mode == NORMAL_CALL)
      req->invoke();   /*method request*/
   else
   {
      req->send_deferred();

      LuaProxy::DeferredClosure* closure = new LuaProxy::DeferredClosure;
      closure->myluaState = L;
      closure->object = this;
      closure->req = req;
      closure->attrib_descr = descr;

      LuaWrapper* DefClosureWrapper = 
      LuaWrapper::getWrapper(L, LuaProxy::wrap_DEFCLOSURE);
      DefClosureWrapper->push(L, closure);
      return 1;
   }
   CORBA_release(req);
   return 0;
}

void LuaProxy::GetAttr(lua_State *L, AttribDescr* descr, InvokeMode mode)
{
   char req_name[70];
   sprintf(req_name,"_get_%s",descr->name);
   CORBA::Request_ptr req = this->orb_object->_request(req_name);

   CORBA::Any* resultAny = descr->mount_result(req);

   if (mode == NORMAL_CALL)
   {
      req->invoke();   /*method request*/

      if (resultAny)
         descr->unmarshal_result(L, resultAny);  // value is returned at
                                                 // the top of the stack
   }
   else
   {
      req->send_deferred();
      LuaProxy::DeferredClosure* closure = new LuaProxy::DeferredClosure;
      closure->myluaState = L;
      closure->object = this;
      closure->req = CORBA::Request::_duplicate(req);
      closure->resultAny = resultAny;
      closure->attrib_descr = descr;

      LuaWrapper* DefClosureWrapper = 
      LuaWrapper::getWrapper(L, LuaProxy::wrap_DEFCLOSURE);
      DefClosureWrapper->push(L, closure);

      return;
   }
   CORBA_release(req);
}


/******************************************************************************/

/**************************** Class LuaProxy::DeferredClosure  *****************/

int LuaProxy::DeferredClosure::Finished(lua_State *L)
{
   try
   {
      VERBOSE_OUT(L, "Deferred Mode: Testing completion")
      if (_finished)
      {
         VERBOSE_OUT(L, "Deferred Mode: Call finished")
         return 1;
      }
      else if (req != NULL && req->poll_response())
      {
         _finished = 1;
         #ifndef ORB_ORBACUS
         req->get_response(); //it is necessary for Visibroker 3.3
         #endif
         VERBOSE_OUT(L, "Deferred Mode: Call finished")
         return 1;
      }
      else
      {
         VERBOSE_OUT(L, "Deferred Mode: Call not finished")
         return 0;
      }
   }
   catch(CORBA::Exception& excep)
   {
      if (method_descr)
         LO_OUT(L) << "Error polling the method " << method_descr->name << endl;
      else
         LO_OUT(L) << "Error polling the attribute " << attrib_descr->name << endl;
      LO_OUT(L) << excep << endl;
   }
   return 0;
}

void LuaProxy::DeferredClosure::Wait(lua_State *L)
{
   try
   {
      if (!_finished && req != NULL)
      {
         VERBOSE_OUT(L, "Deferred Mode: Waiting for response")
         req->get_response();
         _finished = 1;
         VERBOSE_OUT(L, "Deferred Mode: Response received")
      }
   }
   catch(CORBA::Exception& excep)
   {
      if (method_descr)
         LO_OUT(L) << "Error polling the method " << method_descr->name << endl;
      else
         LO_OUT(L) << "Error polling the attribute " << attrib_descr->name << endl;
      LO_OUT(L) << excep << endl;
   }

}
unsigned int LuaProxy::DeferredClosure::GetResult(lua_State *L)
{
   try
   {
      if (_finished && req != NULL)
      {
         VERBOSE_OUT(L, "Deferred Mode: Getting result")

         if (req->env()->exception())
         {
            VERBOSE_OUT(L, "Deferred Mode: Raising an exception")

            return object->RaiseException(L, req->env()->exception(), method_descr);
         }
         else if (resultAny || arguments)
         {
            if (method_descr)
            {
               // unmarshal_result will push the results for Lua
               return method_descr->unmarshal_result(L, resultAny, arguments);
            }
            else
            {
               attrib_descr->unmarshal_result(L, resultAny);
               return 1;
            }
         }
         else
         {
            lua_pushnil(L);
            return 1;
         }
      }
      else
      {
         lua_pushnil(L);
         return 1;
      }
   }
   catch(CORBA::Exception& excep)
   {
      if (method_descr)
         LO_OUT(L) << "Error polling the method " << method_descr->name << endl;
      else
         LO_OUT(L) << "Error polling the attribute " << attrib_descr->name << endl;
      LO_OUT(L) << excep << endl;
      lua_error(L, "LuaOrb: deferred call failed.");
   }
   catch (Err_General& excep)
   {
      lua_error(L, excep.msg);
   }
   return 0;
}

/***************************************************************************/
/*                                                                         */
/*  LuaProxy Constructor                                                   */
/*                                                                         */
/****************************************************************************/

static int _lo_createproxy(lua_State *L)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);

   const char* interfacename = NULL;
   CORBA::Object_ptr orb_obj = CORBA::Object::_nil();
   LuaProxy* proxy = NULL;

   if (lua_isstring(L, 2))
      interfacename = lua_tostring(L, 2);

   if (lua_isuserdata(L, 1))
      orb_obj = CORBA::Object::_duplicate((CORBA::Object_ptr)lua_touserdata(L,1));
#if defined (ORB_VISIBROKER) || defined (ORB_MICO)
   else if (interfacename && (lua_isnil(L, 1) || (lua_isstring(L, 1) &&
            strncmp("name:",lua_tostring(L, 1),5)==0)))
   {
      char* name = NULL;
      if (lua_isstring(L, 1))
      {
         name = lua_tostring(L, 1);
         name = &name[5];
      }
      orb_obj = loCfg->orb->bind(interfacename,name);
   }
#endif
   else if (lua_isstring(L, 1))
   {
      const char* ior = lua_tostring(L, 1);
      try
      {
         orb_obj = loCfg->orb->string_to_object(ior);
      }
      catch (CORBA::Exception& e)
      {
         luaL_verror(L,"Invalid IOR: %s",exception_name(e));
      }
   }
   else
   {
      lua_pushnil(L);
      return 1;
   }

   if (!CORBA::is_nil(orb_obj))
   {
      try {
         // if (LuaOrbCfg::verbose) dump_info(orb_obj);
         proxy = new LuaProxy(L, interfacename, orb_obj);
      }
      catch (Err_General& e)
      {
         luaL_verror(L, "LuaOrb cannot create a proxy.\n%s", e.msg);
      }
      catch (...)
      {
         lua_error(L, "LuaOrb cannot create a proxy.");
      }
   }

   if (proxy != NULL)
      LuaProxy::push(L, proxy);
   else
      lua_pushnil(L);
   return 1;
}

/***************************************************************************/
/*                                                                         */
/*  Tag-Methods and Functions related to LuaOrb Proxies                    */
/*                                                                         */
/****************************************************************************/
static int _callmethod(lua_State *L)
{
   unsigned int nresult = 0;

   /* upvalues : proxy tag, method description and invoke mode */
   int TAG = (int) lua_tonumber(L, -3);
   MethodDescr* method_descr = (MethodDescr*) lua_touserdata(L, -2);
   InvokeMode mode = (InvokeMode) (int) lua_tonumber(L, -1);

   LuaProxy* proxy = LuaProxy::check_tag(L, 1, TAG);

    /* Method parameters are in the Lua stack, beginning at stack position 2 */
   try
   {
      unsigned int nparams = lua_gettop(L) - 3;  /* 2 upvalues + proxy */
      nresult = proxy->Call(L, 2, nparams, method_descr, mode);
   }
   catch (Err_General& excep)
   {
      lua_error(L, excep.msg);
   }
   catch (Err_NParams& excep)
   {
      luaL_verror(L, "%s: invalid number of parameters\n %i arguments and %i formal params",
                      method_descr->name, excep.args, excep.formal);
   }
   catch (Err_Param& excep)
   {
      luaL_verror(L, "%s: the argument %d couldn't be coerced to %s",
                     method_descr->name, excep.index, excep.type);
   }
   catch (Err_NotImpl& excep)
   {
      luaL_verror(L, "%s: the %s type isn't implemented", method_descr->name, excep.type);
   }
   catch(CORBA::Exception& excep)
   {
      LO_OUT(L) << "Error calling the method " << method_descr->name << endl;
      LO_OUT(L) << excep << endl;
   }
   return nresult;
}

/*
** Proxy table is arg #1, index is arg #2
*/
static int index_fb (lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   //the interface repository is required
   if (lua_isstring(L, 2) && (lua_tag(L, 1) == TAG))
   {
      LuaProxy* proxy = LuaProxy::from_lua(L, 1);
      InvokeMode mode;

      try
      {
      char* name = (char*) lua_tostring(L, 2);
      name = LuaProxy::parsename(name,&mode);
      MethodDescr* descr = proxy->myclass->get_opdef(L, name);
      if (descr != NULL)
      {
         lua_pushnumber(L, TAG);
         lua_pushuserdata(L, descr);
         lua_pushnumber(L, (float)mode);
         lua_pushcclosure(L, (lua_CFunction)_callmethod, 3);
      }
      else
      {
         // it is not an operation
         // now, we will check if it is an attribute
         AttribDescr* descr = proxy->myclass->get_attdef(name);
         if (descr != NULL)
         {
            try
            {
               proxy->GetAttr(L, descr, mode);
            }
            catch (Err_NotImpl& excep)
            {
               luaL_verror(L, "attribute %s: the %s type isn't implemented",
                           descr->name,excep.type);
            }
            catch(CORBA::Exception& excep)
            {
               LO_OUT(L) << "Error getting the attribute " << descr->name << endl;
               LO_OUT(L) << excep << endl;
            }
         }
         else
         {
            // it does not belong to the interface
            lua_pushnil(L);
         }
      }
     }
     catch (Err_General& excep)
     {
        lua_error(L, excep.msg);
     }
   }
   else
      lua_pushnil(L);
   return 1;
}

/*
** Proxy table is arg #1, index is arg #2, value is arg #3
*/

static int settable_fb(lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   //the interface repository is required
   if (lua_isstring(L, 2) && (lua_tag(L, 1) == TAG))
   {
      LuaProxy* proxy = LuaProxy::from_lua(L, 1);
      InvokeMode mode;
      int nresult = 0;

      try
      {
      char* attrib_name = (char*) lua_tostring(L, 2);
      attrib_name = LuaProxy::parsename(attrib_name, &mode);
      AttribDescr* descr = proxy->myclass->get_attdef(attrib_name);
      if (descr != NULL)
      {
         if (descr->orb_descr->mode == CORBA(ATTR_READONLY))
         {
            luaL_verror(L,"attribute %s is read-only", descr->name);
            return 0;
         }
         try
         {
            nresult = proxy->SetAttr(L, 3, descr, mode);
         }
         catch (Err_Param& excep)
         {
            luaL_verror(L,"attribute %s: couldn't be coerced to %s",
                          descr->name,excep.type);
         }
         catch (Err_NotImpl& excep)
         {
            luaL_verror(L, "attribute %s: the %s type isn't implemented",
                           descr->name,excep.type);
         }
         catch(CORBA::Exception& excep)
         {
            LO_OUT(L) << "Error setting the attribute " << descr->name << endl;
            LO_OUT(L) << excep << endl;
         }
         return nresult;
      }
      }
      catch (Err_General& excep)
      {
         lua_error(L, excep.msg);
      }
   }

   /* If not an interface attribute, set it locally:
      table, index and value are in the stack */
   lua_pushvalue(L, 2);
   lua_pushvalue(L, 3);
   lua_rawset(L, 1);
   return 0;
}

static int gc_fb(lua_State *L)
{
   LuaProxy* proxy = (LuaProxy*) lua_touserdata(L, 1);
   if ( ! proxy->locked )
   {
      proxy->DiscardLuaThis(L);
      delete proxy;
      VERBOSE_OUT(L, "object collected")
   }
   return 0;
}

/***************************************************************************/
/*                                                                         */
/*  Closure's Tag-Methods                                                  */
/*                                                                         */
/****************************************************************************/
static int defclosure_gc_fb(lua_State *L)
{
   LuaProxy::DeferredClosure* closure = (LuaProxy::DeferredClosure*) lua_touserdata(L,1);
   delete closure;

   VERBOSE_OUT(L, "closure collected")
   return 0;
}

/***************************************************************************/
/*                                                                         */
/*  Closure's Methods                                                      */
/*                                                                         */
/****************************************************************************/
static int defclosure_finished(lua_State *L)
{
   LuaWrapper* DefClosureWrapper = LuaWrapper::getWrapper(L, LuaProxy::wrap_DEFCLOSURE);
   LuaProxy::DeferredClosure* closure = 
     (LuaProxy::DeferredClosure*) DefClosureWrapper->check_tag(L, 1);
   int result = closure->Finished(L);
   if (result)
      lua_pushnumber(L, 1.0);
   else
      lua_pushnil(L);
   return 1;
}

static int defclosure_wait(lua_State *L)
{
   LuaWrapper* DefClosureWrapper = LuaWrapper::getWrapper(L, LuaProxy::wrap_DEFCLOSURE);
   LuaProxy::DeferredClosure* closure = 
     (LuaProxy::DeferredClosure*) DefClosureWrapper->check_tag(L, 1);
   closure->Wait(L);
   return 0;
}

static int defclosure_getresult(lua_State *L)
{
   LuaWrapper* DefClosureWrapper = LuaWrapper::getWrapper(L, LuaProxy::wrap_DEFCLOSURE);
   LuaProxy::DeferredClosure* closure = 
     (LuaProxy::DeferredClosure*) DefClosureWrapper->check_tag(L, 1);

   if (closure->Finished(L))
   {
      //It pushs the method's results in the Lua stack
      return closure->GetResult(L);
   }
   else
   {
      lua_error(L, "LuaOrb: deferred call did not finish.");
      return 0;
   }
}

/***************************************************************************/
/*                                                                         */
/*  Predefined Operations of LuaOrb Proxies                                */
/*                                                                         */
/****************************************************************************/
static int _get_ior(lua_State *L)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);

   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   LuaProxy* proxy  = LuaProxy::check_tag(L, 1, TAG);

   char* ior = loCfg->orb->object_to_string(proxy->orb_object);
   lua_pushstring(L, ior);
   CORBA::string_free(ior);
   return 1;
}

static int _get_repository_id(lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   LuaProxy* proxy  = LuaProxy::check_tag(L, 1, TAG);

   if (proxy->myclass->repository_id())
      lua_pushstring(L, proxy->myclass->repository_id());
   else
      lua_pushnil(L);
   return 1;
}

static int _get_interface_name(lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   LuaProxy* proxy  = LuaProxy::check_tag(L, 1, TAG);

   if (proxy->myclass->interface_name())
      lua_pushstring(L, proxy->myclass->interface_name());
   else
      lua_pushnil(L);
   return 1;
}

static int _get_interface_def(lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   LuaProxy* proxy  = LuaProxy::check_tag(L, 1, TAG);
   LuaProxy* newproxy = NULL;

   CORBA::InterfaceDef_ptr itf_def = proxy->myclass->get_interfacedef();
   if (itf_def)
      newproxy = new LuaProxy(L, "CORBA::InterfaceDef", itf_def);

   if (newproxy)
   {
      newproxy->locked = 1;
      LuaProxy::push(L, newproxy);
   }
   else
      lua_pushnil(L);
   return 1;
}

static int _get_corba_ref(lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   LuaProxy* proxy  = LuaProxy::check_tag(L, 1, TAG);

   lua_pushuserdata(L, proxy->orb_object);
   return 1;
}

static int _non_existent(lua_State *L)
{
   // Proxy tag is provided as an upvalue
   int TAG = (int) lua_tonumber(L, -1);

   LuaProxy* proxy  = LuaProxy::check_tag(L, 1, TAG);

   try
   {
   if ((proxy->orb_object)->_non_existent()) 
      lua_pushnumber(L, (float)1.0);
   else
      lua_pushnil(L);
   }
   catch (...)
   {
      lua_pushnumber(L, (float)1.0);
   }
   return 1;
}

/*****************************************************************************/


static struct luaL_reg functions_tb []= {
   {"lo_createproxy",(lua_CFunction)_lo_createproxy}
};

static struct luaL_reg defclosure_tagmethods_tb[]= {
   {"gc", (lua_CFunction)defclosure_gc_fb}
};

static struct luaL_reg defclosure_methods_tb[]= {
   {"Finished", (lua_CFunction)defclosure_finished},
   {"Wait",(lua_CFunction)defclosure_wait},
   {"GetResult",(lua_CFunction)defclosure_getresult}
};

static char luaorb_code[] =
"-----------------------------------------------------------------------\n"
"-- Function to narrow an object reference to its more specialized type \n"
"-----------------------------------------------------------------------\n"
"function lo_narrow(obj,interface)\n"
"   return lo_createproxy(obj:_get_corba_ref(),interface)\n"
"end";

void LuaProxy::open(lua_State *L)
{
   int TAG = lua_newtag(L);
   lua_pushnumber(L, TAG);
   lua_setglobal(L,"LUAORB_TAG");

   int UDTAG = lua_newtag(L);

   lua_pushnumber(L, TAG);
   lua_pushcclosure(L, (lua_CFunction)index_fb, 1);
   lua_settagmethod(L, TAG, "index");

   lua_pushnumber(L, TAG);
   lua_pushcclosure(L, (lua_CFunction)settable_fb, 1);
   lua_settagmethod(L, TAG,"settable");

   lua_pushcfunction(L, (lua_CFunction)gc_fb);
   lua_settagmethod(L, UDTAG,"gc");

   LuaWrapper* DefClosureWrapper = LuaWrapper::createWrapper(L, LuaProxy::wrap_DEFCLOSURE);
   DefClosureWrapper->reg_ud_tagmethods(L, defclosure_tagmethods_tb,
                      sizeof(defclosure_tagmethods_tb)/sizeof(struct luaL_reg));
   DefClosureWrapper->reg_methods(defclosure_methods_tb,
                      sizeof(defclosure_methods_tb)/sizeof(struct luaL_reg));


   luaL_openlib(L, functions_tb,sizeof(functions_tb)/sizeof(struct luaL_reg));

   // Proxy tags are stored in Lua Registry
   lua_getregistry(L);
   lua_pushstring(L, LuaProxy::reg_TAG);
   lua_pushnumber(L, TAG);
   lua_settable(L, -3);
   lua_pushstring(L, LuaProxy::reg_UDTAG);
   lua_pushnumber(L, UDTAG);
   lua_settable(L, -3);
   lua_pop(L, 1);

   lua_dostring(L, luaorb_code);
}

void LuaProxy::close(lua_State *L)
{
   LuaWrapper::deleteWrapper(L, LuaProxy::wrap_DEFCLOSURE);
}
